"""
Contains the main class that controls the state of the app
"""


import curses
import re
import signal
import time
from math import ceil
from types import FrameType
from typing import Callable, List, Optional, Tuple, Union

from logria.commands.regex import reset_regex_status
from logria.communication.input_handler import InputStream
from logria.communication.render import determine_position
from logria.communication.setup import setup_streams
from logria.interface import color_handler
from logria.interface.textbox import Textbox, rectangle
from logria.logger.parser import Parser
from logria.logger.processor import process_matches, process_parser
from logria.utilities import constants
from logria.utilities.keystrokes import resolve_keypress, validator
from logria.utilities.regex_generator import get_real_length


class Logria():
    """
    Main app class that controls the logical flow of the app
    """

    def __init__(self, stream: Optional[InputStream], history_tape_cache: bool = True, smart_poll_rate: bool = True, poll_rate=0.001):
        # UI Elements initialized to None
        # The entire window
        self.stdscr: curses.window = None  # type: ignore
        # The entire window
        self.outwin: curses.window = None  # type: ignore
        # The command line
        self.command_line: curses.window = None  # type: ignore
        self.box: Textbox  # The text box inside the command line

        # App state passed as parameters
        # Wether we save command history
        self.history_tape_cache: bool = history_tape_cache
        self.poll_rate: float = poll_rate  # The rate at which we check for new messages
        # Wehether we reduce the poll rate to the message receive speed
        self.smart_poll_rate: bool = smart_poll_rate

        # App state that changes as we use the app
        self.first_run: bool = True  # Whether this is a first run or not
        self.height: int = 0  # Window height
        self.width: int = 0  # Window width
        self.loop_time: float = 0  # How long a loop of the main app takes
        # Store the state of the previous render so we know if we need to refresh
        self.previous_render: Optional[Tuple[int, int]] = None
        # Pointer to the previous non-parsed message list, which is continuously updated
        self.previous_messages: List[str] = []
        self.exit_val = 0  # If exit_val is -1, the app dies

        # Message buffers
        self.stderr_messages: List[str] = []
        self.stdout_messages: List[str] = []
        # Default to watching stderr
        self.messages: List[str] = self.stderr_messages

        # Regex Handler information
        # Regex func that handles filtering
        self.func_handle: Optional[Callable] = None
        self.regex_pattern: str = ''  # Current regex pattern
        # List of matches when filtering is active
        self.matched_rows: List[int] = []
        self.last_index_regexed: int = 0  # The last index the filtering function saw

        # Processor information
        self.parser: Optional[Parser] = None  # Reference to the current parser
        self.parser_index: int = 0  # Index for the parser to look at
        self.parsed_messages: List[dict] = []  # List of parsed rows
        self.analytics_enabled: bool = False  # List for statistics messages
        self.last_index_processed: int = 0  # The last index the parsing function saw

        # Variables to store the current state of the app
        self.insert_mode: bool = False  # Default to insert mode (like vim) off
        self.current_status: str = ''  # Current status, aka what is in the command line
        # Determines whether we highlight the match to the user
        self.highlight_match: bool = True
        self.last_row: int = 0  # The last row we can render, aka number of lines
        self.stick_to_bottom: bool = True  # Whether we should follow the stream
        # Whether we should stick to the top and not render new lines
        self.stick_to_top: bool = False
        self.manually_controlled_line: bool = False  # Whether manual scroll is active
        self.current_end: int = 0  # Current last row we have rendered

        # If we do not have a stream yet, tell the user to set one up
        if stream is None:
            self.streams: List[InputStream] = []
        else:
            # Stream list to handle multiple streams
            self.streams = [stream]

    def build_command_line(self) -> None:
        """
        Creates a textbox object that represents the command line
        """
        if self.command_line:
            del self.command_line
        height, width = self.stdscr.getmaxyx()
        # 1 line, screen width, start 2 from the bottom, 1 char from the side
        self.command_line = curses.newwin(1, width, height - 2, 1)
        # Do not block the event loop waiting for input
        self.command_line.nodelay(True)
        # Draw rectangle around the command line
        # upper left:  (height - 2, 0), 2 chars up on left edge
        # lower right: (height, width), bottom right corner of screen
        rectangle(self.stdscr, height - 3, 0, height - 1, width - 2)
        self.stdscr.refresh()
        # Editable text box element
        self.box = Textbox(
            self.command_line,
            insert_mode=self.insert_mode,
            poll_rate=self.poll_rate,
            history_cache=self.history_tape_cache)
        self.write_to_command_line(
            self.current_status)  # Update current status

    def render_text_in_output(self) -> None:
        """
        Renders stream content in the output window

        If filters are inactive, we use `messages`. If they are active, we pull from `matched_rows`

        We write the whole message, regardless of length, because slicing a string allocates a new string
        """
        # Store a pointer to the buffer of messages
        if self.func_handle is None:
            messages_pointer = self.messages
        else:
            # Ignore typing because we use different values depending on what this pointer is
            messages_pointer = self.matched_rows  # type: ignore

        # Determine the start and end position of the render
        start, end = determine_position(self, messages_pointer)
        # Don't do anything if nothing changed; start at index 0
        if not self.analytics_enabled and self.previous_render == (max(start, 0), end):
            return  # Early escape
        self.previous_render = (max(start, 0), end)
        self.outwin.erase()
        current_row = self.last_row  # The row we are currently rendering
        for i in range(end, start, -1):
            if messages_pointer is self.messages:
                # No processing needed for normal messages
                item = messages_pointer[i]
            elif messages_pointer is self.matched_rows:
                # Grab the matched message and optionally highlight it
                messages_idx = self.matched_rows[i]
                item = self.messages[messages_idx]
                if self.highlight_match:
                    # Remove all color codes before applying highlighter
                    item = re.sub(constants.ANSI_COLOR_PATTERN, '', item)
                    match = re.search(self.regex_pattern, item)
                    if match:
                        start, end = match.span()
                        matched_str = item[start:end]
                        item = item.replace(
                            matched_str, f'\u001b[35m{matched_str}\u001b[0m')
            # Find the correct start position
            current_row -= ceil(get_real_length(item) / self.width)
            if current_row < 0:
                break
            # Instead of window.addstr, handle colors
            color_handler.addstr(self.outwin, current_row, 0, item.rstrip())
        self.outwin.refresh()

    def write_to_command_line(self, string: str) -> None:
        """
        Writes a message to the command line
        """
        self.reset_command_line()
        curses.curs_set(1)
        self.command_line.move(0, 0)
        self.command_line.addstr(0, 0, string)
        curses.curs_set(0)

    def reset_command_line(self) -> None:
        """
        Resets the command line
        """
        self.command_line.move(0, 0)
        self.command_line.deleteln()
        curses.curs_set(0)

    def activate_prompt(self, text='') -> None:
        """
        Activate the prompt so we can edit it

        text: str, some text to prepend to the command
        """
        self.reset_command_line()
        if text:
            self.write_to_command_line(text)
        curses.curs_set(1)
        self.box.edit(validator)

    def update_poll_rate(self, new_poll_rate: float) -> None:
        """
        Update Logria's poll rate
        """
        self.poll_rate = new_poll_rate
        # Update stream poll rates
        for stream in self.streams:
            stream.poll_rate = new_poll_rate
        # Update command line poll rate
        self.box.poll_rate = new_poll_rate

    def handle_smart_poll_rate(self, t_1: float, new_messages: int) -> None:
        """
        Determine a reasonable poll rate based on the speed of messages received
        """
        if self.manually_controlled_line:
            pass
        elif self.smart_poll_rate:
            if not self.loop_time:
                self.loop_time = time.perf_counter()
            else:
                self.loop_time = t_1 - self.loop_time
                messages_per_second = new_messages / self.loop_time
                if messages_per_second > 0:
                    # Update poll rate
                    new_poll_rate = \
                        min(
                            max(
                                1 / messages_per_second,
                                constants.FASTEST_POLL_RATE
                            ),
                            constants.SLOWEST_POLL_RATE
                        )
                    self.update_poll_rate(new_poll_rate)

    def redraw(self) -> None:
        """
        Force re-render the content of the window as if it has never been rendered before
        """
        self.previous_render = None  # Force render
        self.render_text_in_output()

    def resize_window(self) -> None:
        """
        Resize curses elements when window size changes
        """
        self.height, self.width = self.stdscr.getmaxyx()
        self.width -= 1  # Leave space for padding on the right
        self.last_row = self.height - 3  # The last row we can write to
        curses.resizeterm(self.height, self.width + 1)
        self.build_command_line()  # Rebuild the command line
        # Re-reate the window with these sizes
        self.outwin = curses.newwin(
            self.height - 3, self.width, 0, 0)
        self.redraw()

    def start(self) -> None:
        """
        Starts the program
        """
        signal.signal(signal.SIGINT, self.stop)  # type: ignore
        curses.wrapper(self.main)

    # pylint: disable=unused-argument
    def stop(self, signum: Optional[Union[str, int]] = None, frame: Optional[FrameType] = None) -> None:
        """
        Die if we send an exit signal of -1
        """
        for stream in self.streams:
            stream.exit()
        self.exit_val = -1
        # If we crash before the command line is set up:
        if getattr(self, 'box', None):
            self.box.stop()

    def main(self, stdscr) -> None:
        """
        Main program loop, handles user control and logical flow
        """
        curses.use_default_colors()
        self.stdscr = stdscr
        stdscr.keypad(1)
        height, width = stdscr.getmaxyx()  # Get screen size

        # Save these values
        self.height = height
        self.width = width - 1

        # Setup Output window
        output_start_row = 0  # Leave space for top border
        output_height = height - 3  # Leave space for command line
        self.last_row = output_height - output_start_row  # The last row we can write to
        # Create the window with these sizes
        self.outwin = curses.newwin(
            output_height, width - 1, output_start_row, 0)
        self.outwin.refresh()

        # Setup Command line
        self.build_command_line()

        # Update the command line status
        reset_regex_status(self)

        # Disable cursor:
        curses.curs_set(0)

        # If the streams do not exist, create them
        if not self.streams:
            setup_streams(self)

        # Start the main app loop
        while True:
            # Update messages from the input stream's queues, track time
            t_0 = time.perf_counter()
            new_messages: int = 0
            for stream in self.streams:
                while not stream.stderr.empty():
                    message = stream.stderr.get()
                    self.stderr_messages.append(message)
                    new_messages += 1

                while not stream.stdout.empty():
                    message = stream.stdout.get()
                    self.stdout_messages.append(message)
                    new_messages += 1
            # Prevent this loop from taking up 100% of the CPU dedicated to the main thread by delaying loops
            t_1 = time.perf_counter() - t_0
            # Don't delay if the queue processing took too long
            time.sleep(max(0, self.poll_rate - t_1))

            # Calculate new poll rate
            if self.smart_poll_rate:
                self.handle_smart_poll_rate(t_1, new_messages)

            # Since this is the first run, set the visible stream to the one that has the most messages
            if self.first_run and (len(self.stdout_messages) > 0 or len(self.stderr_messages) > 0):
                self.first_run = False
                # Default to stdout unless stderr has more messages
                if len(self.stdout_messages) >= len(self.stderr_messages):
                    self.messages = self.stdout_messages
                else:
                    self.messages = self.stderr_messages

            try:
                # Get keypress, raise curses.error if nothing detected
                keypress = self.command_line.getkey()
                resolve_keypress(self, keypress)
            except curses.error:
                if self.exit_val == -1:
                    return
                # If we have an active filter/parser, process it/them
                if self.parser:
                    # This may block if there are a lot of messages
                    process_parser(self)
                if self.func_handle:
                    # This may block if there are a lot of messages
                    process_matches(self)
                # Always try to render
                self.render_text_in_output()



"""
Logria window content display handler
"""


from logria.commands.regex import reset_regex_status
from logria.commands.parser import reset_parser
# from logria.communication.shell_output import Logria


def swap_input(logria: 'Logria'):  # type: ignore
    """
    Swap between stderr and stdout
    """
    logria.previous_render = None  # Force render, defer draw
    # Swap stdout and stderr
    reset_parser(logria)
    reset_regex_status(logria)
    if logria.messages is logria.stderr_messages:
        logria.messages = logria.stdout_messages
    else:
        logria.messages = logria.stderr_messages


def resize(logria: 'Logria'):  # type: ignore
    """
    Resize window
    """
    logria.resize_window()



"""
Unit Tests for sessions
"""

import unittest
from os import remove
from pathlib import Path

from logria.utilities import session
from logria.utilities.constants import SAVED_SESSIONS_PATH


class TestSessionHandler(unittest.TestCase):
    """
    Test cases to ensure constants are the expected values
    """

    def test_can_setup_handler(self):
        """
        Test that we can create an instance of the class
        """
        s = session.SessionHandler()
        self.assertIsInstance(s, session.SessionHandler)

    def test_setup_folder(self):
        """
        Ensure we correctly setup the folder in the expected place
        """
        s = session.SessionHandler()
        s.setup_folder()
        self.assertEqual(s.folder, Path(Path.home(), SAVED_SESSIONS_PATH))

    def test_sessions(self):
        """
        Test that we can get the sessions as a dict
        """
        s = session.SessionHandler()
        actual = s.sessions()
        expected = {
            0: 'Cmd - Generate Test Logs',
            1: 'File - Sample Access Log',
            2: 'File - readme'
        }
        self.assertEqual(actual, expected)

    def test_show_sessions(self):
        """
        Test that we can get the sessions as a list
        """
        s = session.SessionHandler()
        actual = s.show_sessions()
        expected = ['0: Cmd - Generate Test Logs',
                    '1: File - Sample Access Log',
                    '2: File - readme'
                    ]
        self.assertEqual(actual, expected)

    def test_session_parsing(self):
        """
        Test that we properly show sessions from the directory as a list
        """
        s = session.SessionHandler()
        d = s.sessions()
        l = s.show_sessions()
        self.assertEqual([f'{i}: {d[i]}' for i in d], l)

    def test_load_session(self):
        """
        Test that both ways we load sessions are the same
        """
        s = session.SessionHandler()
        sessions = s.sessions()
        first_item = list(sessions.keys())[0]
        actual = s.load_session(first_item)
        expected = \
            {
                "commands": [
                    [
                        "/Users/chris/Documents/Code/Python/logria/venv/bin/python3",
                        "/Users/chris/.logria/sample_streams/generate_test_logs.py"
                    ],
                    [
                        "/Users/chris/Documents/Code/Python/logria/venv/bin/python3",
                        "/Users/chris/.logria/sample_streams/generate_test_logs_2.py"
                    ]
                ],
                "type": "command"
            }
        self.assertEqual(actual, expected)

    def test_save_session(self):
        """
        Test that we can write a session if we need to
        """
        s = session.SessionHandler()
        s.save_session('Test', ['ls', '-l'], 'cmd')
        remove(s.folder / 'Test')

    def test_save_current_session(self):
        """
        Test that we can write a session after programatic creation
        """
        s = session.SessionHandler()
        sessions = s.sessions()
        first_item = list(sessions.keys())[0]
        s.load_session(first_item)
        s.save_current_session('Test Session')
        self.assertIn('Test Session', s.sessions().values())
        s.remove_session('Test Session')
        self.assertNotIn('Test Session', s.sessions().values())

    def test_remove_session(self):
        """
        Test that we can write a session if we need to
        """
        s = session.SessionHandler()
        s.save_session('Test', ['ls', '-l'], 'cmd')
        s.remove_session('Test')
        self.assertNotIn('Test', s.sessions().values())

    def test_cannot_remove_session(self):
        """
        Test that we can write a session if we need to
        """
        s = session.SessionHandler()
        self.assertFalse(s.remove_session('Test'))


class TestSessionBuilder(unittest.TestCase):
    """
    Tests the constructor methods of a SessionHander
    """

    def test_set_params(self):
        """
        Test that we can set a command's params
        """
        s = session.SessionHandler()
        command = ['ls', '-l']
        type_ = 'command'
        s.set_params(command, type_)
        self.assertEqual(s._commands, [command])
        self.assertEqual(s._type, type_)

    def test_as_list_command(self):
        """
        Test that we can resolve a list of commands
        """
        s = session.SessionHandler()
        command = ['ls', '-l']
        command_2 = ['grep', '-nr', 'code', '.']
        type_ = 'command'
        s.set_params(command, type_)
        s.add_command(command_2)
        expected = ['Type:', '  command',
                    'Commands:', '  ls -l', '  grep -nr code .']
        self.assertEqual(s.as_list(), expected)

    def test_as_list_file(self):
        """
        Test that we can resolve a list of files
        """
        s = session.SessionHandler()
        command = ['usr', 'log', 'cloud-init.log']
        command_2 = ['usr', 'log', 'cloud-init-output.log']
        type_ = 'file'
        s.set_params(command, type_)
        s.add_command(command_2)
        expected = ['Type:', '  file', 'Files:',
                    '  usr/log/cloud-init.log', '  usr/log/cloud-init-output.log']
        self.assertEqual(s.as_list(), expected)



"""
Tests the app launcher
"""

import os
import unittest

from logria.commands import scroll
from logria.communication.render import determine_position
from logria.communication.shell_output import Logria
from logria.logger.processor import process_matches
from logria.utilities import regex_generator
from logria.utilities.keystrokes import resolve_keypress


class TestCanRenderContentRange(unittest.TestCase):
    """
    Tests scenarios with which we render content
    """

    def test_can_render_matches(self):
        """
        Test we render properly when using matches
        """
        os.environ['TERM'] = 'dumb'
        app = Logria(None, False, False)

        # Fake window size: 10 x 100
        app.height = 10
        app.width = 100

        # Set fake previous render
        app.last_row = app.height - 3  # simulate the last row we can render to
        app.current_end = 80  # Simulate the last message rendered

        # Set fake messages
        app.messages = [f'{x}|{x}' for x in range(20)]

        # Set fake filter
        app.func_handle = regex_generator.regex_test_generator(r'\d\|\d')
        process_matches(app)

        # Set positional booleans
        app.manually_controlled_line = False
        app.stick_to_top = True
        app.stick_to_bottom = False

        start, end = determine_position(app, app.matched_rows)
        self.assertEqual(start, -1)
        self.assertEqual(end, 6)
        app.stop()

    def test_render_final_items(self):
        """
        Test we render properly when stuck to the bottom
        """
        os.environ['TERM'] = 'dumb'
        app = Logria(None, False, False)

        # Fake window size: 10 x 100
        app.height = 10
        app.width = 100

        # Set fake previous render
        app.last_row = app.height - 3  # simulate the last row we can render to
        app.current_end = 80  # Simulate the last message rendered

        # Set fake messages
        app.messages = [str(x) for x in range(100)]

        # Set positional booleans
        app.manually_controlled_line = False
        app.stick_to_top = False
        app.stick_to_bottom = True

        start, end = determine_position(app, app.messages)
        self.assertEqual(start, 91)
        self.assertEqual(end, 99)
        app.stop()

    def test_render_first_items(self):
        """
        Test we render properly when stuck to the top
        """
        os.environ['TERM'] = 'dumb'
        app = Logria(None, False, False)

        # Fake window size: 10 x 100
        app.height = 10
        app.width = 100

        # Set fake previous render
        app.last_row = app.height - 3  # simulate the last row we can render to
        app.current_end = 80  # Simulate the last message rendered

        # Set fake messages
        app.messages = [str(x) for x in range(100)]

        # Set positional booleans
        app.manually_controlled_line = False
        app.stick_to_top = True
        app.stick_to_bottom = False

        start, end = determine_position(app, app.messages)
        self.assertEqual(start, -1)
        self.assertEqual(end, 6)
        app.stop()

    def test_render_few_items_from_middle(self):
        """
        Test we render properly when stuck to the top
        """
        os.environ['TERM'] = 'dumb'
        app = Logria(None, False, False)

        # Fake window size: 10 x 100
        app.height = 10
        app.width = 100

        # Set fake previous render
        app.last_row = app.height - 3  # simulate the last row we can render to
        app.current_end = 2  # Simulate the last message rendered

        # Set fake messages
        app.messages = [str(x) for x in range(4)]

        # Set positional booleans
        app.manually_controlled_line = True
        app.stick_to_top = False
        app.stick_to_bottom = False

        start, end = determine_position(app, app.messages)
        self.assertEqual(start, -1)
        self.assertEqual(end, 3)
        app.stop()

    def test_render_middle_items(self):
        """
        Test we render properly when stuck to the top
        """
        os.environ['TERM'] = 'dumb'
        app = Logria(None, False, False)

        # Fake window size: 10 x 100
        app.height = 10
        app.width = 100

        # Set fake previous render
        app.last_row = app.height - 3  # simulate the last row we can render to
        app.current_end = 80  # Simulate the last message rendered

        # Set fake messages
        app.messages = [str(x) for x in range(100)]

        # Set positional booleans
        app.manually_controlled_line = True
        app.stick_to_top = False
        app.stick_to_bottom = False

        start, end = determine_position(app, app.messages)
        self.assertEqual(start, 72)
        self.assertEqual(end, 80)
        app.stop()

    def test_render_middle_items_early(self):
        """
        Test we render properly when stuck to the top
        """
        os.environ['TERM'] = 'dumb'
        app = Logria(None, False, False)

        # Fake window size: 10 x 100
        app.height = 10
        app.width = 100

        # Set fake previous render
        app.last_row = app.height - 3  # simulate the last row we can render to
        app.current_end = 5  # Simulate the last message rendered

        # Set fake messages
        app.messages = [str(x) for x in range(100)]

        # Set positional booleans
        app.manually_controlled_line = True
        app.stick_to_top = False
        app.stick_to_bottom = False

        start, end = determine_position(app, app.messages)
        self.assertEqual(start, -1)
        self.assertEqual(end, 5)
        app.stop()

    def test_render_small_list_from_top(self):
        """
        Test we render properly when stuck to the top
        """
        os.environ['TERM'] = 'dumb'
        app = Logria(None, False, False)

        # Fake window size: 10 x 100
        app.height = 10
        app.width = 100

        # Set fake previous render
        app.last_row = app.height - 3  # simulate the last row we can render to
        app.current_end = 0  # Simulate the last message rendered

        # Set fake messages
        app.messages = [str(x) for x in range(5)]

        # Set positional booleans
        app.manually_controlled_line = False
        app.stick_to_top = True
        app.stick_to_bottom = False

        start, end = determine_position(app, app.messages)
        self.assertEqual(start, -1)
        self.assertEqual(end, 4)
        app.stop()

    def test_render_small_list_from_bottom(self):
        """
        Test we render properly when stuck to the top
        """
        os.environ['TERM'] = 'dumb'
        app = Logria(None, False, False)

        # Fake window size: 10 x 100
        app.height = 10
        app.width = 100

        # Set fake previous render
        app.last_row = app.height - 3  # simulate the last row we can render to
        app.current_end = 0  # Simulate the last message rendered

        # Set fake messages
        app.messages = [str(x) for x in range(5)]

        # Set positional booleans
        app.manually_controlled_line = False
        app.stick_to_top = False
        app.stick_to_bottom = True

        start, end = determine_position(app, app.messages)
        self.assertEqual(start, -1)
        self.assertEqual(end, 4)
        app.stop()

    def test_render_scroll_up(self):
        """
        Test we render properly when stuck to the bottom
        """
        os.environ['TERM'] = 'dumb'
        app = Logria(None, False, False)

        # Fake window size: 10 x 100
        app.height = 10
        app.width = 100

        # Set fake previous render
        app.last_row = app.height - 3  # simulate the last row we can render to
        app.current_end = 99  # Simulate the last message rendered

        # Set fake messages
        app.messages = [str(x) for x in range(100)]

        # Set positional booleans
        app.manually_controlled_line = False
        app.stick_to_top = False
        app.stick_to_bottom = True

        # Scroll action
        scroll.up(app)

        start, end = determine_position(app, app.messages)
        self.assertEqual(start, 90)
        self.assertEqual(end, 98)
        app.stop()

    def test_render_scroll_down(self):
        """
        Test we render properly when stuck to the bottom
        """
        os.environ['TERM'] = 'dumb'
        app = Logria(None, False, False)

        # Fake window size: 10 x 100
        app.height = 10
        app.width = 100

        # Set fake previous render
        app.last_row = app.height - 3  # simulate the last row we can render to
        app.current_end = 10  # Simulate the last message rendered

        # Set fake messages
        app.messages = [str(x) for x in range(100)]

        # Set positional booleans
        app.manually_controlled_line = False
        app.stick_to_top = True
        app.stick_to_bottom = True

        # Scroll action
        scroll.down(app)

        start, end = determine_position(app, app.messages)
        self.assertEqual(start, 3)
        self.assertEqual(end, 11)
        app.stop()

    def test_render_scroll_down_matched(self):
        """
        Test we render properly when stuck to the bottom
        """
        os.environ['TERM'] = 'dumb'
        app = Logria(None, False, False)

        # Fake window size: 10 x 100
        app.height = 10
        app.width = 100

        # Set fake previous render
        app.last_row = app.height - 3  # simulate the last row we can render to
        app.current_end = 10  # Simulate the last message rendered

        # Set fake messages
        app.messages = [str(x) for x in range(100)]
        app.matched_rows = [x for x in range(20)]

        # Set positional booleans
        app.manually_controlled_line = False
        app.stick_to_top = True
        app.stick_to_bottom = True

        # Scroll action
        scroll.down(app)

        start, end = determine_position(app, app.messages)
        self.assertEqual(start, 3)
        self.assertEqual(end, 11)
        app.stop()

    def test_render_stick_bottom(self):
        """
        Test we render properly when stuck to the bottom
        """
        os.environ['TERM'] = 'dumb'
        app = Logria(None, False, False)

        # Fake window size: 10 x 100
        app.height = 10
        app.width = 100

        # Set fake previous render
        app.last_row = app.height - 3  # simulate the last row we can render to
        app.current_end = 80  # Simulate the last message rendered

        # Set fake messages
        app.messages = [str(x) for x in range(100)]

        # Set positional booleans
        app.manually_controlled_line = False
        app.stick_to_top = True
        app.stick_to_bottom = False

        # Scroll action
        scroll.bottom(app)

        start, end = determine_position(app, app.messages)
        self.assertEqual(start, 91)
        self.assertEqual(end, 99)
        app.stop()

    def test_render_scroll_past_end(self):
        """
        Test we render properly when stuck to the bottom
        """
        os.environ['TERM'] = 'dumb'
        app = Logria(None, False, False)

        # Fake window size: 10 x 100
        app.height = 10
        app.width = 100

        # Set fake previous render
        app.last_row = app.height - 3  # simulate the last row we can render to
        app.current_end = 101  # Simulate the last message rendered

        # Set fake messages
        app.messages = [str(x) for x in range(100)]

        # Set positional booleans
        app.manually_controlled_line = True
        app.stick_to_top = False
        app.stick_to_bottom = False

        start, end = determine_position(app, app.messages)
        self.assertEqual(start, 92)
        self.assertEqual(end, 100)
        app.stop()

    def test_render_everything_unset(self):
        """
        Test we render properly when stuck to the bottom
        """
        os.environ['TERM'] = 'dumb'
        app = Logria(None, False, False)

        # Fake window size: 10 x 100
        app.height = 10
        app.width = 100

        # Set fake previous render
        app.last_row = app.height - 3  # simulate the last row we can render to
        app.current_end = 101  # Simulate the last message rendered

        # Set fake messages
        app.messages = [str(x) for x in range(100)]

        # Set positional booleans
        app.manually_controlled_line = False
        app.stick_to_top = False
        app.stick_to_bottom = False

        start, end = determine_position(app, app.messages)
        self.assertEqual(start, 92)
        self.assertEqual(end, 100)
        app.stop()

    def test_render_stick_top(self):
        """
        Test we render properly when stuck to the top
        """
        os.environ['TERM'] = 'dumb'
        app = Logria(None, False, False)

        # Fake window size: 10 x 100
        app.height = 10
        app.width = 100

        # Set fake previous render
        app.last_row = app.height - 3  # simulate the last row we can render to
        app.current_end = 80  # Simulate the last message rendered

        # Set fake messages
        app.messages = [str(x) for x in range(100)]

        # Set positional booleans
        app.manually_controlled_line = False
        app.stick_to_top = False
        app.stick_to_bottom = True

        # Scroll action
        scroll.top(app)

        start, end = determine_position(app, app.messages)
        self.assertEqual(start, -1)
        self.assertEqual(end, 6)
        app.stop()

    def test_render_scroll_pgdn(self):
        """
        Test we render properly when stuck to the bottom
        """
        os.environ['TERM'] = 'dumb'
        app = Logria(None, False, False)

        # Fake window size: 10 x 100
        app.height = 10
        app.width = 100

        # Set fake previous render
        app.last_row = app.height - 3  # simulate the last row we can render to
        app.current_end = 10  # Simulate the last message rendered

        # Set fake messages
        app.messages = [str(x) for x in range(100)]

        # Set positional booleans
        app.manually_controlled_line = False
        app.stick_to_top = True
        app.stick_to_bottom = True

        # Scroll action
        scroll.pgdn(app)

        start, end = determine_position(app, app.messages)
        self.assertEqual(start, 9)
        self.assertEqual(end, 17)
        app.stop()

    def test_render_scroll_pgup(self):
        """
        Test we render properly when stuck to the bottom
        """
        os.environ['TERM'] = 'dumb'
        app = Logria(None, False, False)

        # Fake window size: 10 x 100
        app.height = 10
        app.width = 100

        # Set fake previous render
        app.last_row = app.height - 3  # simulate the last row we can render to
        app.current_end = 40  # Simulate the last message rendered

        # Set fake messages
        app.messages = [str(x) for x in range(100)]

        # Set positional booleans
        app.manually_controlled_line = False
        app.stick_to_top = True
        app.stick_to_bottom = True

        # Scroll action
        scroll.pgup(app)

        start, end = determine_position(app, app.messages)
        self.assertEqual(start, 25)
        self.assertEqual(end, 33)
        app.stop()


class TestCanRenderContentRangeKeypress(unittest.TestCase):
    """
    Tests scenarios with which we render content
    """

    def test_render_scroll_up_keypress(self):
        """
        Test we render properly when stuck to the bottom
        """
        os.environ['TERM'] = 'dumb'
        app = Logria(None, False, False)

        # Fake window size: 10 x 100
        app.height = 10
        app.width = 100

        # Set fake previous render
        app.last_row = app.height - 3  # simulate the last row we can render to
        app.current_end = 99  # Simulate the last message rendered

        # Set fake messages
        app.messages = [str(x) for x in range(100)]

        # Set positional booleans
        app.manually_controlled_line = False
        app.stick_to_top = False
        app.stick_to_bottom = True

        # Scroll action
        resolve_keypress(app, 'KEY_UP')

        start, end = determine_position(app, app.messages)
        self.assertEqual(start, 90)
        self.assertEqual(end, 98)
        app.stop()

    def test_render_scroll_down_keypress(self):
        """
        Test we render properly when stuck to the bottom
        """
        os.environ['TERM'] = 'dumb'
        app = Logria(None, False, False)

        # Fake window size: 10 x 100
        app.height = 10
        app.width = 100

        # Set fake previous render
        app.last_row = app.height - 3  # simulate the last row we can render to
        app.current_end = 10  # Simulate the last message rendered

        # Set fake messages
        app.messages = [str(x) for x in range(100)]

        # Set positional booleans
        app.manually_controlled_line = False
        app.stick_to_top = True
        app.stick_to_bottom = True

        # Scroll action
        resolve_keypress(app, 'KEY_DOWN')

        start, end = determine_position(app, app.messages)
        self.assertEqual(start, 3)
        self.assertEqual(end, 11)
        app.stop()

    def test_render_stick_bottom_keypress(self):
        """
        Test we render properly when stuck to the bottom
        """
        os.environ['TERM'] = 'dumb'
        app = Logria(None, False, False)

        # Fake window size: 10 x 100
        app.height = 10
        app.width = 100

        # Set fake previous render
        app.last_row = app.height - 3  # simulate the last row we can render to
        app.current_end = 80  # Simulate the last message rendered

        # Set fake messages
        app.messages = [str(x) for x in range(100)]

        # Set positional booleans
        app.manually_controlled_line = False
        app.stick_to_top = True
        app.stick_to_bottom = False

        # Scroll action
        resolve_keypress(app, 'KEY_RIGHT')

        start, end = determine_position(app, app.messages)
        self.assertEqual(start, 91)
        self.assertEqual(end, 99)
        app.stop()

    def test_render_stick_top_keypress(self):
        """
        Test we render properly when stuck to the top
        """
        os.environ['TERM'] = 'dumb'
        app = Logria(None, False, False)

        # Fake window size: 10 x 100
        app.height = 10
        app.width = 100

        # Set fake previous render
        app.last_row = app.height - 3  # simulate the last row we can render to
        app.current_end = 80  # Simulate the last message rendered

        # Set fake messages
        app.messages = [str(x) for x in range(100)]

        # Set positional booleans
        app.manually_controlled_line = False
        app.stick_to_top = False
        app.stick_to_bottom = True

        # Scroll action
        resolve_keypress(app, 'KEY_LEFT')

        start, end = determine_position(app, app.messages)
        self.assertEqual(start, -1)
        self.assertEqual(end, 6)
        app.stop()

    def test_render_scroll_pgdn_keystroke(self):
        """
        Test we render properly when stuck to the bottom
        """
        os.environ['TERM'] = 'dumb'
        app = Logria(None, False, False)

        # Fake window size: 10 x 100
        app.height = 10
        app.width = 100

        # Set fake previous render
        app.last_row = app.height - 3  # simulate the last row we can render to
        app.current_end = 10  # Simulate the last message rendered

        # Set fake messages
        app.messages = [str(x) for x in range(100)]

        # Set positional booleans
        app.manually_controlled_line = False
        app.stick_to_top = True
        app.stick_to_bottom = True

        # Scroll action
        resolve_keypress(app, 'KEY_NPAGE')

        start, end = determine_position(app, app.messages)
        self.assertEqual(start, 9)
        self.assertEqual(end, 17)
        app.stop()

    def test_render_scroll_pgup_keystroke(self):
        """
        Test we render properly when stuck to the bottom
        """
        os.environ['TERM'] = 'dumb'
        app = Logria(None, False, False)

        # Fake window size: 10 x 100
        app.height = 10
        app.width = 100

        # Set fake previous render
        app.last_row = app.height - 3  # simulate the last row we can render to
        app.current_end = 40  # Simulate the last message rendered

        # Set fake messages
        app.messages = [str(x) for x in range(100)]

        # Set positional booleans
        app.manually_controlled_line = False
        app.stick_to_top = True
        app.stick_to_bottom = True

        # Scroll action
        resolve_keypress(app, 'KEY_PPAGE')

        start, end = determine_position(app, app.messages)
        self.assertEqual(start, 25)
        self.assertEqual(end, 33)
        app.stop()



"""
Unit Tests for setup functions
"""

import unittest

from logria.commands.config import resolve_delete_command


class SetupValidator(unittest.TestCase):
    """
    Test cases to ensure setup operates correctly
    """

    def test_resolve_single_num(self):
        """
        Test that we correctly validate 1 singleton
        """
        resolved = resolve_delete_command(':r 1')
        self.assertCountEqual(resolved, [1])

    def test_resolve_double_num(self):
        """
        Test that we correctly a 2 singletons
        """
        resolved = resolve_delete_command(':r 1,2')
        self.assertCountEqual(resolved, [1, 2])

    def test_resolve_triple_num(self):
        """
        Test that we correctly validate 3 singletons
        """
        resolved = resolve_delete_command(':r 1,2,3')
        self.assertCountEqual(resolved, [1, 2, 3])

    def test_resolve_triple_num_trailing_comma(self):
        """
        Test that we correctly validate 3 singletons with trailing comma
        """
        resolved = resolve_delete_command(':r 1,2,3,')
        self.assertCountEqual(resolved, [1, 2, 3])

    def test_resolve_range(self):
        """
        Test that we correctly validate 1 range
        """
        resolved = resolve_delete_command(':r 1-5')
        self.assertCountEqual(resolved, [1, 2, 3, 4, 5])

    def test_resolve_double_range(self):
        """
        Test that we correctly validate two ranges
        """
        resolved = resolve_delete_command(':r 1-3,5-7')
        self.assertCountEqual(resolved, [1, 2, 3, 5, 6, 7])

    def test_resolve_triple_range(self):
        """
        Test that we correctly validate three ranges
        """
        resolved = resolve_delete_command(':r 1-3,5-7,9-11')
        self.assertCountEqual(resolved, [1, 2, 3, 5, 6, 7, 9, 10, 11])

    def test_resolve_ranges_with_singletons(self):
        """
        Test that we correctly validate ranges with singletons
        """
        resolved = resolve_delete_command(':r 1-3,5,9-11,15')
        self.assertCountEqual(resolved, [1, 2, 3, 5, 9, 10, 11, 15])

    def test_resolve_ranges_multiple_dash(self):
        """
        Test that we correctly validate range with multiple dashes
        """
        resolved = resolve_delete_command(':r 1--3,4')
        self.assertCountEqual(resolved, [4])

    def test_resolve_ranges_with_string(self):
        """
        Test that we correctly delete with range
        """
        resolved = resolve_delete_command(':r a-b,4')
        self.assertCountEqual(resolved, [4])



"""
Module for enumerating tools on the PATH so that we may use
fully qualified names when spawning new shells
"""


import os
from pathlib import Path
from typing import List

from logria.logger.default_logger import setup_default_logger

# Setup default logger
LOGGER = setup_default_logger(__name__)


class Resolver():
    """
    Resolves a tool to its fully qualified name if it exists on the user's PATH
    """

    def __init__(self):
        self._paths = {}
        self.user_home = str(Path.home())
        self.resolve_paths()

    def resolve_paths(self) -> None:
        """
        Fill the paths dictionary with data
        """
        paths = os.environ.get('PATH', None)
        if paths is None:
            LOGGER.info('PATH environment variable does not exist!')
        else:
            # Iterate in reverse so we resolve tools in local paths after system paths
            for path in paths.split(':')[::-1]:
                # Resolve home dir symbols
                full_path = self.resolve_home_dir(path)
                # Try and remove the file from the path to include the parent dir
                if os.path.isfile(full_path):
                    print(
                        f'{full_path} listed in PATH environment variable, refers to file!')
                    full_path = '/'.join(full_path.split('/')[:-1])
                try:
                    # No need to use fs wrapper here
                    programs = os.listdir(self.resolve_home_dir(full_path))
                except FileNotFoundError:
                    print(
                        f'{full_path} listed in PATH environment variable, but does not exist!')
                    continue
                self._paths.update(
                    {programs: f'{full_path}/{programs}' for programs in programs}
                )

    def get(self, program: str) -> str:
        """
        Get a program's fully qualified path if it exists
        """
        return self._paths.get(program, program)

    def resolve_home_dir(self, part: str) -> str:
        """
        Resolves the user's home directory
        """
        if '~' in part:
            return part.replace('~', self.user_home)
        elif '$HOME' in part:
            return part.replace('$HOME', self.user_home)
        return part

    def resolve_command_as_string(self, command: str) -> str:
        """
        Replace all programs on the path with their fully qualified counterparts as a string
        """
        return ' '.join(self.resolve_command_as_list(command))

    def resolve_command_as_list(self, command: str) -> List[str]:
        """
        Replace all programs on the path with their fully qualified counterparts as a list
        """
        command_parts = command.split(' ')
        new_command = []
        for part in command_parts:
            resolved_part = self.get(part)
            resolved_part = self.resolve_home_dir(resolved_part)
            new_command.append(resolved_part)
        return new_command

    def resolve_file_as_list(self, filepath: str) -> List[str]:
        """
        Resolve a file path as a list of folders
        """
        path_parts = filepath.split('/')
        new_path = []
        for part in path_parts:
            new_path.append(self.resolve_home_dir(part))
        return new_path

    def resolve_file_as_str(self, filepath: str) -> str:
        """
        Resolve a file path as a slash-separated string
        """
        return '/'.join(self.resolve_file_as_list(filepath))



"""
Commands for enabling parser and analytics
"""


import time
from json import JSONDecodeError

from logria.commands.config import resolve_delete_command
from logria.commands.regex import reset_regex_status
from logria.logger.parser import Parser

# from logria.communication.shell_output import Logria


def reset_parser(logria: 'Logria', custom_message: str = ''):  # type: ignore
    """
    Remove the current parser, if any exists
    """
    if logria.func_handle:
        logria.current_status = f'Regex with pattern /{logria.regex_pattern}/'
    elif custom_message:
        logria.current_status = custom_message
    else:
        logria.current_status = 'No filter applied'  # CLI message, rendered after
    if logria.previous_messages:
        # Move messages pointer to the previous state
        if logria.previous_messages is logria.stderr_messages:
            logria.messages = logria.stderr_messages
        else:
            logria.messages = logria.stdout_messages
        logria.previous_messages = []
        logria.parsed_messages = []  # Dump parsed messages
    logria.parser = None  # Dump the parser
    logria.analytics_enabled = False  # Disable analytics blocker
    logria.parser_index = 0  # Dump the pattern index
    logria.last_index_processed = 0  # Reset the last searched index
    logria.current_end = 0  # We now do not know where to end
    logria.stick_to_bottom = True  # Stay at the bottom for the next render
    logria.stick_to_top = False  # Do not stick to the top
    logria.manually_controlled_line = False  # Do not stop rendering new messages
    logria.write_to_command_line(logria.current_status)


def setup_parser(logria: 'Logria'):  # type: ignore
    """
    Setup a parser object in the main runtime
    """
    # Reset the status for new writes
    reset_parser(logria)
    reset_regex_status(logria)

    # Store previous message pointer
    if logria.messages is logria.stderr_messages:
        logria.previous_messages = logria.stderr_messages
    elif logria.messages is logria.stdout_messages:
        logria.previous_messages = logria.stdout_messages

    # Stick to top to show options
    logria.manually_controlled_line = False
    logria.stick_to_bottom = True
    logria.stick_to_top = False

    # Only set when there are no pattenrs to display so we can update the user
    custom_message = ''

    # Create parser object
    parser = Parser()

    # Overwrite the messages pointer
    logria.messages = parser.show_patterns()
    logria.redraw()
    while True:
        if not parser.show_patterns():
            parser = None  # type: ignore
            custom_message = 'No parsers found! Enter :config to build one. Press z to cancel.'
            break
        time.sleep(logria.poll_rate)
        logria.activate_prompt()
        command = logria.box.gather().strip()
        if command == ':q':
            logria.messages = logria.previous_messages
            parser = None  # type: ignore
            break
        if ':r ' in command[:3]:
            items_to_remove = resolve_delete_command(command)
            sessions = parser.patterns()
            for item in items_to_remove:
                if item in sessions:
                    parser.remove(sessions[item])
            logria.messages = parser.show_patterns()
            logria.redraw()
            continue
        try:
            parser.load(parser.patterns()[int(command)])
            break
        except KeyError:
            continue
        except JSONDecodeError as err:
            logria.messages.append(
                f'Invalid JSON: {err.msg} on line {err.lineno}, char {err.colno}')
        except ValueError:
            pass

    # Ensure we didn't exit early already
    if parser:
        # Overwrite a different list this time, and reset it when done
        logria.messages = parser.display_example()
        logria.previous_render = None
        logria.render_text_in_output()
        while True:
            time.sleep(logria.poll_rate)
            logria.activate_prompt()
            command = logria.box.gather().strip()
            if command == ':q':
                logria.messages = logria.previous_messages
                logria.previous_render = None
                reset_parser(logria)
                return  # no need to do any other work
            try:
                command = int(command)
                if command > len(logria.messages):
                    raise ValueError(f'Invalid index {command} not in range 0..{len(logria.messages)}')
                logria.parser_index = int(command)
                logria.current_status = f'Parsing with {parser.get_name()}, field {parser.get_analytics_for_index(command)}'
                logria.write_to_command_line(logria.current_status)
                break
            except ValueError:
                pass
            except AssertionError:
                pass
        # Set parser
        logria.parser = parser

        # Put pointer back to new list
        logria.messages = logria.parsed_messages
    else:
        logria.messages = logria.previous_messages
        reset_parser(logria, custom_message=custom_message)

    # Render immediately
    logria.previous_render = None

    # Stick to bottom again
    logria.last_index_processed = 0
    logria.stick_to_bottom = True
    logria.stick_to_top = False


def enable_parser(logria: 'Logria'):  # type: ignore
    """
    Enable parser
    """
    if logria.parser is not None:
        reset_parser(logria)
    setup_parser(logria)


def enable_analytics(logria: 'Logria'):  # type: ignore
    """
    Enable analytics engine
    """
    if logria.parser is not None:
        logria.last_index_processed = 0
        logria.parser.reset_analytics()
        if logria.analytics_enabled:
            logria.current_status = f'Parsing with {logria.parser.get_name()}, field {logria.parser.get_analytics_for_index(logria.parser_index)}'
            logria.parsed_messages = []
            logria.analytics_enabled = False
        else:
            logria.analytics_enabled = True
            logria.current_status = f'Parsing with {logria.parser.get_name()}, analytics view'


def teardown_parser(logria: 'Logria'):  # type: ignore
    """
    Tear down parser
    """
    reset_parser(logria)



"""
Commands for scrolling the window
"""


# from logria.communication.shell_output import Logria
from logria.utilities import constants

def pgup(logria: 'Logria') -> None:  # type: ignore
    """
    Handle page up keypress
    """
    for _ in range(logria.last_row):
        up(logria)


def pgdn(logria: 'Logria') -> None:  # type: ignore
    """
    Handle page down keypress
    """
    for _ in range(logria.last_row):
        down(logria)


def up(logria: 'Logria') -> None:  # type: ignore
    """
    Scroll one line up
    """
    # Smooth scroll
    if logria.smart_poll_rate:
        logria.update_poll_rate(constants.FASTEST_POLL_RATE)
    # Scroll up
    logria.manually_controlled_line = True
    logria.stick_to_top = False
    logria.stick_to_bottom = False
    logria.current_end = max(0, logria.current_end - 1)
    logria.previous_render = None  # Force render, defer draw


def down(logria: 'Logria'):  # type: ignore
    """
    Scroll one line down
    """
    # Smooth scroll
    if logria.smart_poll_rate:
        logria.update_poll_rate(constants.FASTEST_POLL_RATE)
    # Scroll down
    logria.manually_controlled_line = True
    logria.stick_to_top = False
    logria.stick_to_bottom = False
    if logria.matched_rows:
        logria.current_end = min(
            len(logria.matched_rows) - 1, logria.current_end + 1)
    else:
        logria.current_end = min(
            len(logria.messages) - 1, logria.current_end + 1)
    logria.previous_render = None  # Force render, defer draw


def bottom(logria: 'Logria') -> None:  # type: ignore
    """
    Stick to bottom
    """
    logria.stick_to_top = False
    logria.stick_to_bottom = True
    logria.manually_controlled_line = False


def top(logria: 'Logria') -> None:  # type: ignore
    """
    Stick to top
    """
    logria.stick_to_top = True
    logria.stick_to_bottom = False
    logria.manually_controlled_line = False



"""
Tests the app launcher
"""

import os
import unittest
from curses import error

from logria.communication.input_handler import (CommandInputStream,
                                                FileInputStream)
from logria.communication.shell_output import Logria


class TestCanLaunchApp(unittest.TestCase):
    """
    Tests various ways we can launch the app
    """

    def test_launch_no_stream(self):
        """
        Test that we can launch the app
        This should crash only due to a curses error
        """
        with self.assertRaises(error):
            os.environ['TERM'] = 'dumb'
            app = Logria(None, False)
            app.start()

    def test_launch_with_command_stream(self):
        """
        Test that we can launch the app
        This should crash only due to a curses error
        """
        with self.assertRaises(error):
            os.environ['TERM'] = 'dumb'
            stream = CommandInputStream(['ls', '-l'])
            app = Logria(stream, False)
            app.start()

    def test_launch_with_file_stream(self):
        """
        Test that we can launch the app
        This should crash only due to a curses error
        """
        with self.assertRaises(error):
            os.environ['TERM'] = 'dumb'
            stream = FileInputStream(['readme.md'])
            app = Logria(stream, False)
            app.start()



"""
Unit Tests for command parser
"""

import os
import unittest
from pathlib import Path

from logria.utilities import command_parser


class TestFormattingFunctions(unittest.TestCase):
    """
    Test cases to ensure command parser are the expected values
    """

    def test_can_create_resolver(self):
        """
        Test that we can create a resolver instance
        """
        r = command_parser.Resolver()
        self.assertIsInstance(r, command_parser.Resolver)

    def test_can_get_user_home(self):
        """
        Test that we get the user's home folder properly
        """
        r = command_parser.Resolver()
        self.assertEqual(r.user_home, str(Path.home()))

    def test_can_resolve_real_path(self):
        """
        Test that we correctly resolve a real path
        """
        r = command_parser.Resolver()
        self.assertEqual(r.get('tail'), '/usr/bin/tail')

    def test_can_resolve_fake_path(self):
        """
        Test that we correctly resolve a fake path
        """
        r = command_parser.Resolver()
        self.assertEqual(r.get('tail123'), 'tail123')

    def test_resolve_home_dir(self):
        """
        Test that we resolve the proper strings to the user's home directory
        """
        r = command_parser.Resolver()
        expected = str(Path.home())
        self.assertEqual(r.resolve_home_dir('~'), expected)
        self.assertEqual(r.resolve_home_dir('$HOME'), expected)

    def test_resolve_command_as_string(self):
        """
        Test that we properly resolve a command to a string
        """
        r = command_parser.Resolver()
        actual = r.resolve_command_as_string(
            'tail -f $HOME/file.txt | cat -ne')
        expected = f'/usr/bin/tail -f {Path.home()}/file.txt | /bin/cat -ne'
        self.assertEqual(actual, expected)

    def test_resolve_command_as_list(self):
        """
        Test that we properly resolve a command to a list
        """
        r = command_parser.Resolver()
        actual = r.resolve_command_as_list(
            'tail -f ~/file.txt | cat -ne')
        expected = [
            '/usr/bin/tail',
            '-f',
            f'{Path.home()}/file.txt',
            '|',
            '/bin/cat',
            '-ne'
        ]
        self.assertEqual(actual, expected)

    def test_resolve_file_as_string(self):
        """
        Test that we properly resolve a file to a string
        """
        r = command_parser.Resolver()
        actual = r.resolve_file_as_str('$HOME/file.txt')
        expected = f'{Path.home()}/file.txt'
        self.assertEqual(actual, expected)

    def test_resolve_file_as_list(self):
        """
        Test that we properly resolve a file to a list
        """
        r = command_parser.Resolver()
        actual = r.resolve_file_as_list(
            '~/file.txt')
        expected = [
            f'{Path.home()}',
            'file.txt'
        ]
        self.assertEqual(actual, expected)


class TestParserPathEnv(unittest.TestCase):
    """
    Tests whether we correctly resolve PATH variables
    """
    def test_no_path(self):
        """
        Test we do not crash if the path varibale is missing
        """
        path_string = os.environ.pop("PATH")
        r = command_parser.Resolver()
        self.assertFalse(r._paths)
        # Put the PATH back for other tests
        os.environ["PATH"] = path_string

    def test_path_folder_missing(self):
        """
        Test we do not crash if the folder does not exist
        """
        os.environ['PATH'] += ':/a/a/b/b/c/c/d/d'
        r = command_parser.Resolver()
        self.assertTrue(r._paths)

    def test_file_on_path(self):
        """
        Test we do not crash if there is a file on the path, make sure it gets added
        """
        os.environ['PATH'] += ':~/.logria/sessions/.DS_Store'
        r = command_parser.Resolver()
        self.assertIn('.DS_Store', r._paths)



"""
Logria editing bar commands
"""


# from logria.communication.shell_output import Logria


def toggle_insert_mode(logria: 'Logria'):  # type: ignore
    """
    Toggle insert mode
    """
    if logria.insert_mode:
        logria.insert_mode = False
    else:
        logria.insert_mode = True
    logria.build_command_line()



"""
Constants used throughout the application
"""

import os
from pathlib import Path

# Patterns
ANSI_COLOR_PATTERN = r'(\x9B|\x1B\[)[0-?]*[ -\/]*[@-~]'

# Paths
USER_HOME = '' if os.environ.get('LOGRIA_DISABLE_USER_HOME') else str(Path.home())
LOGRIA_ROOT = os.environ.get('LOGRIA_ROOT', '.logria')
SAVED_PATTERNS_PATH = f'{USER_HOME}/{LOGRIA_ROOT}/patterns'
SAVED_SESSIONS_PATH = f'{USER_HOME}/{LOGRIA_ROOT}/sessions'
SAVED_HISTORY_PATH = f'{USER_HOME}/{LOGRIA_ROOT}/history'

# Filenames
HISTORY_TAPE_NAME = 'tape'

# Numerical limits
FASTEST_POLL_RATE: float = 0.0001   # Fast enough for smooth typing, 1000 hz
SLOWEST_POLL_RATE: float = 0.1  # Poll ten times per second, 10 hz

# Text to exclude from message history
HISTORY_EXCLUDES = {
    ':history',
    ':history off'
}

# Messages
START_MESSAGE = [
    'Enter a new command to open and save a new stream,',
    'or enter a number to choose a saved session from the list,',
    'or enter `:config` to configure.',
    ' ',  # Blank line for printout
    'Enter `:r #` to remove session #.',
    'Enter `:q` to quit.',
    ' '  # Not an empty string so Curses knows to not use this line
]

# Config messages
CONFIG_START_MESSAGES = [
    'Saved data paths:',
    f'Parsers:  {USER_HOME}/{SAVED_PATTERNS_PATH}',
    f'Sessions: {USER_HOME}/{SAVED_SESSIONS_PATH}',
    'To configure new parameters, enter `session` or `parser`'
]
CREATE_SESSION_START_MESSAGES = [
    'To create a session, enter a type, either `command` or `file`:'
]
CREATE_PARSER_MESSAGES = [
    'To create a parser, enter a type, either `regex` or `split`:'
]

# Session Strings
SESSION_ADD_COMMAND = 'Enter a command to open pipes to:'
SESSION_SHOULD_CONTINUE_COMMAND = 'Enter :s to save or press enter to add another command'
SESSION_ADD_FILE = 'Enter a path to a file:'
SESSION_SHOULD_CONTINUE_FILE = 'Enter :s to save or press enter to add another file'
SAVE_CURRENT_SESSION = 'Enter a name to save the session:'

# Parser Strings
PARSER_SET_NAME = 'Enter a name for the parser:'
PARSER_SET_EXAMPLE = 'Enter an example string to match against:'
PARSER_SET_PATTERN = 'Enter a regex pattern:'
SAVE_CURRENT_PATTERN = 'Press enter to save or type `:q` to quit:'

# Startup messages
APP_DESCRIPTION = 'A powerful CLI tool that puts log analytics at your fingertips.'
EXEC_HELP = 'Command to listen to, ex: logria -e \'tail -f log.txt\''
HISTORY_HELP = 'Disable command history disk cache'
SMART_SPEED_HELP = 'Disable variable speed polling based on message receive rate'
PIPE_INPUT_ERROR = \
'''Piping is not supported as Logria cannot both
listen to stdin as well as get user input from
your tty. Process substitution is also not
allowed, as Logria is unable to read from the
file descriptor created by the shell.

To capture command output, start Logria and
enter the command during the setup process, or
invoke Logria with `logria -e \'command\''''



"""
Handles parsing color escape code sequences in logs to the relevant curses
colors so that we do not get ugly strings like `033[94m Blue`

Adapted from https://github.com/spellr/culour/blob/master/culour/culour.py
"""

import os
import curses
from typing import Dict, Tuple

COLOR_PAIRS_CACHE: Dict[Tuple[int, int], int] = {}
DEFAULT_COLOR = -1


# Source: https://en.wikipedia.org/wiki/ANSI_escape_code#3/4_bit
FOREGROUND = {
    # Normal
    '[37': curses.COLOR_WHITE,
    '[36': curses.COLOR_CYAN,
    '[35': curses.COLOR_MAGENTA,
    '[34': curses.COLOR_BLUE,
    '[33': curses.COLOR_YELLOW,
    '[32': curses.COLOR_GREEN,
    '[31': curses.COLOR_RED,
    '[30': curses.COLOR_BLACK,
    # Bright
    '[97': curses.COLOR_WHITE,
    '[96': curses.COLOR_CYAN,
    '[95': curses.COLOR_MAGENTA,
    '[94': curses.COLOR_BLUE,
    '[93': curses.COLOR_YELLOW,
    '[92': curses.COLOR_GREEN,
    '[91': curses.COLOR_RED,
    '[90': curses.COLOR_BLACK,
}
BACKGROUND = {
    # Normal
    '[47': curses.COLOR_WHITE,
    '[46': curses.COLOR_CYAN,
    '[45': curses.COLOR_MAGENTA,
    '[44': curses.COLOR_BLUE,
    '[43': curses.COLOR_YELLOW,
    '[42': curses.COLOR_GREEN,
    '[41': curses.COLOR_RED,
    '[40': curses.COLOR_BLACK,
    # Bright
    '[107': curses.COLOR_WHITE,
    '[106': curses.COLOR_CYAN,
    '[105': curses.COLOR_MAGENTA,
    '[104': curses.COLOR_BLUE,
    '[103': curses.COLOR_YELLOW,
    '[102': curses.COLOR_GREEN,
    '[101': curses.COLOR_RED,
    '[100': curses.COLOR_BLACK,
}


def _get_color(foreground: int, background: int):
    """
    Memoize init_pair wrapper; store an index of previously used color combinations

    init_color creates an integer and stores it as the primary key to a color pair,
    so here we cache the results so we do not overwrite old pairs
    """
    key = (foreground, background)
    if key not in COLOR_PAIRS_CACHE:
        # Use the pairs from 101 and after, so there's less chance they'll be overwritten by the user
        pair_num = len(COLOR_PAIRS_CACHE) + 101
        try:
            curses.init_pair(pair_num, foreground, background)
        except curses.error:
            # If colors were never enabled, this call does not matter anyway
            pass
        COLOR_PAIRS_CACHE[key] = pair_num

    return COLOR_PAIRS_CACHE[key]


def _color_str_to_color_pair(color: str):
    """
    Convert the escape code color to the curses color binding
    """
    foreground = FOREGROUND.get(color, DEFAULT_COLOR)
    background = BACKGROUND.get(color, DEFAULT_COLOR)
    color_pair = _get_color(foreground, background)
    return color_pair


def _sanitize(line: str) -> str:
    """
    Sanitize null bytes from strings before we try and render them
    """
    if '\x00' in line:
        line = line.replace('\x00', '')
    return line


def _add_line(y_coord: int, x_coord: int, window, line: str):
    # Sanitize string
    line = _sanitize(line)
    # split but \033 which stands for a color change
    color_split = line.split('\033')

    # Print the first part of the line without color change
    default_color_pair = _get_color(DEFAULT_COLOR, DEFAULT_COLOR)
    try:
        window.addstr(y_coord, x_coord,
                      color_split[0], curses.color_pair(default_color_pair))
        window.noutrefresh()
        y_coord, x_coord = curses.getsyx()
    except curses.error:
        pass

    # Iterate over the rest of the line-parts and print them with their colors
    for substring in color_split[1:]:
        color_str = substring.split('m')[0]
        substring = substring[len(color_str)+1:]
        color_pair = _color_str_to_color_pair(color_str)
        try:
            window.addstr(y_coord, x_coord, substring,
                          curses.color_pair(color_pair))
            window.noutrefresh()
            y_coord, x_coord = curses.getsyx()
        except curses.error:
            pass


def _inner_addstr(window, string: str, y_coord=-1, x_coord=-1):
    if not curses.has_colors:
        raise ValueError(
            'Curses was not configured to support colors. Call `curses.start_color()`')

    cur_y, cur_x = window.getyx()
    if y_coord == -1:
        y_coord = cur_y
    if x_coord == -1:
        x_coord = cur_x
    for line in string.split(os.linesep):
        _add_line(y_coord, x_coord, window, line)
        # next line
        y_coord += 1


def addstr(*args):
    """
    Adds the color-formatted string to the given window, in the given coordinates
    To add in the current location, call like this:
        addstr(window, string)
    and to set the location to print the string, call with:
        addstr(window, y, x, string)
    Only use color pairs up to 100 when using this function,
    otherwise you will overwrite the pairs used by this function
    """
    if len(args) != 2 and len(args) != 4:
        raise TypeError("addstr requires 2 or 4 arguments")

    if len(args) == 4:
        window = args[0]
        y_coord = args[1]
        x_coord = args[2]
        string = args[3]
    else:
        window = args[0]
        string = args[1]
        y_coord = -1
        x_coord = -1

    return _inner_addstr(window, string, y_coord, x_coord)



"""
Logria command handler
"""
import curses

from logria.commands.parser import reset_parser
from logria.utilities import constants
from logria.commands.config import config_mode
from logria.communication.setup import setup_streams

# from logria.communication.shell_output import Logria


def start_history_mode(logria: 'Logria', last_n: int) -> None:  # type: ignore
    """
    Swap message pointer to history tape
    """
    # Store previous message pointer
    if logria.messages is logria.stderr_messages:
        logria.previous_messages = logria.stderr_messages
    elif logria.messages is logria.stdout_messages:
        logria.previous_messages = logria.stdout_messages

    # Set new message pointer
    logria.messages = logria.box.history_tape.tail(last_n=last_n)


def handle_command(logria: 'Logria') -> None:  # type: ignore
    """
    Enable command mode
    """
    # Handle smart poll rate
    if logria.smart_poll_rate:
        # Make it smooth to type
        logria.update_poll_rate(constants.FASTEST_POLL_RATE)
    # Handle getting input from the command line for commands
    logria.activate_prompt(':')
    command = logria.box.gather().strip()
    curses.curs_set(0)
    # Since we parse info out of these strings we cannot use a dictionary to store
    #   the callable map like we do in the main app
    if command:
        if command == ':q':
            logria.stop()
        elif command[:5] == ':poll':
            try:
                new_poll_rate = float(command.replace(':poll', ''))
            except ValueError:
                pass
            else:
                logria.update_poll_rate(new_poll_rate)
        elif command[:7] == ':config':
            config_mode(logria)
        elif command[:8] == ':history':
            if command == ':history off':
                reset_parser(logria)
            else:
                try:
                    num_to_get = int(command.replace(':history', ''))
                except ValueError:
                    num_to_get = logria.height  # Default to screen height if no info given
                start_history_mode(logria, num_to_get)
        elif command[:8] == ':restart':
            # Kill all streams
            for stream in logria.streams:
                stream.exit()
            # Remove them from Logria
            logria.streams = []
            # Reset messages buffers
            logria.stderr_messages = []
            logria.stdout_messages = []
            logria.parsed_messages = []
            logria.matched_rows = []
            # Setup new streams
            setup_streams(logria)
    logria.reset_command_line()
    logria.write_to_command_line(logria.current_status)



"""
Unit Tests for keystrokes
"""

import unittest

from logria.utilities import keystrokes


class TestValidator(unittest.TestCase):
    """
    Test cases to ensure keystrokes are set to expected values
    """

    def test_validator(self):
        """
        Test that we correctly validate keystrokes
        """
        for i in range(300):
            if i != 127:
                self.assertEqual(keystrokes.validator(i), i)
            else:
                self.assertEqual(keystrokes.validator(i), 263)



"""
Handle creating input streams for Logria
"""


import time
from json import JSONDecodeError
from os.path import isfile
from typing import List

from logria.commands.config import config_mode, resolve_delete_command
from logria.communication.input_handler import (CommandInputStream,
                                                FileInputStream)
from logria.utilities import constants
from logria.utilities.command_parser import Resolver
from logria.utilities.session import SessionHandler

# from logria.communication.shell_output import Logria


def render_setup_messages(logria: 'Logria', session_handler: SessionHandler) -> None:  # type: ignore
    """
    Render setup messages from session_handler
    """
    # Create a new message list to see
    setup_messages: List[str] = []
    logria.messages = setup_messages
    # Tell the user what we are doing
    setup_messages.extend(constants.START_MESSAGE)
    setup_messages.extend(session_handler.show_sessions())

    # Force re-render
    logria.redraw()


def setup_streams(logria: 'Logria') -> None:  # type: ignore
    """
    When launched without a stream, allow the user to define them for us
    """
    # Setup a SessionHandler and get the existing saved sessions
    session_handler = SessionHandler()
    render_setup_messages(logria, session_handler)

    # Create resolver class to resolve commands
    resolver = Resolver()

    # Get user input
    while True:
        time.sleep(logria.poll_rate)
        logria.activate_prompt()
        command = logria.box.gather().strip()
        if not command:
            continue
        if command == ':q':
            logria.stop()
            break
        if ':r ' in command[:3]:
            items_to_remove = resolve_delete_command(command)
            sessions = session_handler.sessions()
            for item in items_to_remove:
                if item in sessions:
                    session_handler.remove_session(sessions[item])
            render_setup_messages(logria, session_handler)
            continue
        try:
            chosen_item = int(command)
            session = session_handler.load_session(chosen_item)
            if not session:
                continue
            stored_commands = session['commands']
            # Commands need a type
            for stored_command in stored_commands:
                if session.get('type') == 'file':
                    logria.streams.append(FileInputStream(stored_command))
                elif session.get('type') == 'command':
                    logria.streams.append(CommandInputStream(stored_command))
        except KeyError as err:
            logria.messages.append(
                f'Data missing from configuration: {err}')
            logria.render_text_in_output()
            continue
        except JSONDecodeError as err:
            logria.messages.append(
                f'Invalid JSON: {err.msg} on line {err.lineno}, char {err.colno}')
            logria.render_text_in_output()
            continue
        except ValueError:
            if command == ':config':
                config_mode(logria)
                return
            elif command == ':q':
                logria.stop()
            elif isfile(command):
                logria.streams.append(
                    FileInputStream(command.split('/')))
                session_handler.save_session(
                    'File - ' + command.replace('/', '|'), [command.split('/')], 'file')
            else:
                cmd = resolver.resolve_command_as_list(command)
                logria.streams.append(CommandInputStream(cmd))
                session_handler.save_session(
                    'Cmd - ' + command.replace('/', '|'), cmd, 'command')
        break

    # Launch the subprocess
    for stream in logria.streams:
        stream.poll_rate = logria.poll_rate
        stream.start()

    # Set status back to what it was
    logria.write_to_command_line(logria.current_status)

    # Render immediately
    logria.previous_render = None

    # Reset messages
    logria.stderr_messages = []
    logria.messages = logria.stderr_messages



"""
Functions for handling getting input from a file or from stdout/stderr
"""


import multiprocessing
import time
from fcntl import F_GETFL, F_SETFL, fcntl
from os import O_NONBLOCK
from subprocess import PIPE, Popen
from typing import List


class InputStream():
    """
    Spawns a process that will create queues we can read from to get input from a stream
    """

    def __init__(self, args: List[str], poll_rate=0.001) -> None:
        # Poll processes for new messages at this rate
        self.poll_rate = poll_rate

        # Use separate queues for stdout/stderr so we can parse separately
        self.stdout: multiprocessing.Queue = multiprocessing.Queue()
        self.stderr: multiprocessing.Queue = multiprocessing.Queue()

        # Create the child process, have it run in the background
        self.process = multiprocessing.Process(
            target=self.run, args=(args, self.stdout, self.stderr,))
        self.process.name = ' '.join(str(args))

    def start(self):
        """
        Start the process independent of init
        """
        self.process.start()

    def run(self, args: List[str], stdoutq: multiprocessing.Queue, stderrq: multiprocessing.Queue) -> None:
        """
        Called by the process; should put data in stdoutq and/or stderrq
        """
        raise NotImplementedError(
            'Input stream class initialized from parent!')

    def exit(self):
        """
        Kills the process
        """
        if self.process:
            self.process.terminate()


class CommandInputStream(InputStream):
    """
    Read a subprocess command as an input stream
    """

    def __init__(self, args, poll_rate=0.001):
        super().__init__(args, poll_rate=poll_rate)
        self.proc = None

    def run(self, args: List[str], stdoutq: multiprocessing.Queue, stderrq: multiprocessing.Queue) -> None:
        """
        Given a command passed as an array ['python', 'script.py'], open a
        pipe to it and read the contents

        This will not read python print() calls because print does not flush stdout by default,
          this can be enabled with `print('', flush=True)`
        """
        try:
            self.proc = Popen(args, stdout=PIPE, stderr=PIPE, bufsize=1,
                              universal_newlines=True)

            # Un-buffer streams
            stdout_flag = fcntl(self.proc.stdout, F_GETFL)  # type: ignore
            fcntl(self.proc.stdout, F_SETFL, stdout_flag | O_NONBLOCK)  # type: ignore

            stderr_flag = fcntl(self.proc.stderr, F_GETFL)  # type: ignore
            fcntl(self.proc.stderr, F_SETFL, stderr_flag | O_NONBLOCK)  # type: ignore

            while True:
                time.sleep(self.poll_rate)
                # stdout
                stdout_output = self.proc.stdout.readline()  # type: ignore
                if stdout_output:
                    stdoutq.put(stdout_output)

                # stderr
                stderr_output = self.proc.stderr.readline()  # type: ignore
                if stderr_output:
                    stderrq.put(stderr_output)

                # Kill condition
                if stderr_output == '' and stdout_output == '' and self.proc.poll() is not None:
                    break
        except PermissionError:
            stderrq.put(
                f'Permissions error opening handle to command: {"/".join(args)}')
        except FileNotFoundError:
            stderrq.put(
                f'File not found error opening handle to command: {"/".join(args)}')

    def exit(self):
        """
        Kills the process
        """
        if self.proc is not None:
            # Proc may be dead already when we get here
            self.proc.kill()  # Kill piped process
            self.proc.communicate()  # Make sure any final data is not left in the pipe
        if self.process:
            self.process.terminate()  # Kill Python process


class FileInputStream(InputStream):
    """
    Read in a file as a stream
    """

    def run(self, args: List[str], stdoutq: multiprocessing.Queue, _: multiprocessing.Queue) -> None:
        """
        Given a filename, open the file and read the contents
        args: a list of folders to be joined ['Docs', 'file.py'] -> 'Docs/file.py'
        """
        try:
            with open('/'.join(args), 'r') as f_in:
                for line in f_in.readlines():
                    stdoutq.put(line)
        except PermissionError:
            _.put(
                f'Permissions error opening file handle to: {"/".join(args)}')
        except FileNotFoundError:
            stdoutq.put(
                f'File not found error opening handle to command: {"/".join(args)}')
        except OSError:
            stdoutq.put(f'Bad file descriptor: {args}')



"""
Unit Tests for fs module
"""

import unittest
from pathlib import Path

from logria.utilities import fs


class TestListDirectory(unittest.TestCase):
    """
    Test cases to ensure InputStream can be initialized
    """

    def test_can_resolve_dir_str(self):
        """
        Test that we can resolve a string path
        """
        result = fs.listdir('/', {'.DS_Store'})
        self.assertIsInstance(result, list)

    def test_can_resolve_dir_path(self):
        """
        Test that we can resolve a Pathlib path
        """
        result = fs.listdir(Path('/'), {'.DS_Store'})
        self.assertIsInstance(result, list)

    def test_resolve_invalid_path(self):
        """
        Test that we don't crash if the path does not exist
        """
        result = fs.listdir(Path('a/b/c/d/e/f/g'), {'.DS_Store'})
        self.assertListEqual(result, [])

    def test_can_remove_pattern(self):
        """
        Test that pattern removal works
        """
        result = fs.listdir(Path('/'), {'dev'})
        if result is not None:
            self.assertNotIn('dev', result)



"""
Unit Tests for processors
"""

import os
import unittest

from logria.communication.shell_output import Logria
from logria.logger.parser import Parser
from logria.logger.processor import process_matches, process_parser
from logria.utilities import regex_generator


class TestProcessors(unittest.TestCase):
    """
    Test cases to ensure keystrokes are set to expected values
    """

    def test_process_matches(self):
        """
        Test that we correctly process matches
        """
        os.environ['TERM'] = 'dumb'
        app = Logria(None, False, False)

        # Fake window size: 10 x 100
        app.height = 10
        app.width = 100

        # Set fake previous render
        app.last_row = app.height - 3  # simulate the last row we can render to
        app.current_end = 80  # Simulate the last message rendered

        # Set fake messages
        app.messages = [str(x) for x in range(20)]

        # Set regex function, process
        app.func_handle = regex_generator.regex_test_generator(r'\d')
        process_matches(app)

        self.assertEqual(app.messages, [str(x) for x in range(20)])

    def test_process_parser_no_analytics(self):
        """
        Test that we correctly process parser with no analytics
        """
        os.environ['TERM'] = 'dumb'
        app = Logria(None, False, False)

        # Fake window size: 10 x 100
        app.height = 10
        app.width = 100

        # Set fake previous render
        app.last_row = app.height - 3  # simulate the last row we can render to
        app.current_end = 80  # Simulate the last message rendered

        # Set fake messages
        app.messages = [str(x) for x in range(10)]
        app.parser_index = 0
        app.last_index_processed = 0

        # Set parser, activate
        app.parser = Parser()
        app.parser.set_pattern(
            pattern=r'(\d)',
            type_='regex',
            name='Test',
            example='4',
            analytics_methods={
                'Item': 'average'
            }
        )

        # Store previous message pointer
        app.previous_messages = app.messages

        # Process parser
        process_parser(app)

        self.assertEqual(app.messages, [str(x) for x in range(10)])

    def test_process_parser_invalid_index(self):
        """
        Test that we correctly process parser with invalid index
        """
        os.environ['TERM'] = 'dumb'
        app = Logria(None, False, False)

        # Fake window size: 10 x 100
        app.height = 10
        app.width = 100

        # Set fake previous render
        app.last_row = app.height - 3  # simulate the last row we can render to
        app.current_end = 80  # Simulate the last message rendered

        # Set fake messages
        app.messages = [f'{x}+{x}+{x}' for x in range(10)]
        app.parser_index = 3
        app.last_index_processed = 0

        # Set parser, activate
        app.parser = Parser()
        app.parser.set_pattern(
            pattern='\+',
            type_='split',
            name='Test',
            example='a-a',
            analytics_methods={
                'Item 1': 'count',
                'Item 2': 'count'
            }
        )

        # Store previous message pointer
        app.previous_messages = app.messages

        # Process parser
        process_parser(app)

        self.assertEqual(app.messages, [])

    def test_process_parser_analytics_average(self):
        """
        Test that we correctly process parser with average metric
        """
        os.environ['TERM'] = 'dumb'
        app = Logria(None, False, False)

        # Fake window size: 10 x 100
        app.height = 10
        app.width = 100

        # Set fake previous render
        app.last_row = app.height - 3  # simulate the last row we can render to
        app.current_end = 80  # Simulate the last message rendered

        # Set fake messages
        app.messages = [str(x) for x in range(10)]
        app.parser_index = 0
        app.last_index_processed = 0
        app.analytics_enabled = True

        # Set parser, activate
        app.parser = Parser()
        app.parser.set_pattern(
            pattern=r'(\d)',
            type_='regex',
            name='Test',
            example='4',
            analytics_methods={
                'Item': 'average'
            }
        )

        # Set analytics method manually
        app.parser._analytics_map = dict(
            zip(range(len(app.parser._analytics_methods.keys())), app.parser._analytics_methods.keys()))

        # Store previous message pointer
        app.previous_messages = app.messages

        # Process parser
        process_parser(app)

        self.assertEqual(app.messages, [
                         'Item', '  average:\t 4.50', '  count:\t 10.00', '  total:\t 45.00'])

    def test_process_parser_analytics_count(self):
        """
        Test that we correctly process parser with count metric
        """
        os.environ['TERM'] = 'dumb'
        app = Logria(None, False, False)

        # Fake window size: 10 x 100
        app.height = 10
        app.width = 100

        # Set fake previous render
        app.last_row = app.height - 3  # simulate the last row we can render to
        app.current_end = 80  # Simulate the last message rendered

        # Set fake messages
        app.messages = [str(x) for x in range(10)]
        app.parser_index = 0
        app.last_index_processed = 0
        app.analytics_enabled = True

        # Set parser, activate
        app.parser = Parser()
        app.parser.set_pattern(
            pattern=r'(\d)',
            type_='regex',
            name='Test',
            example='4',
            analytics_methods={
                'Item': 'count'
            }
        )

        # Set analytics method manually
        app.parser._analytics_map = dict(
            zip(range(len(app.parser._analytics_methods.keys())), app.parser._analytics_methods.keys()))

        # Store previous message pointer
        app.previous_messages = app.messages

        # Process parser
        process_parser(app)

        self.assertEqual(
            app.messages, ['Item', '  0: 1', '  1: 1', '  2: 1', '  3: 1', '  4: 1'])

    def test_process_parser_analytics_sum(self):
        """
        Test that we correctly process parsers with sum metric
        """
        os.environ['TERM'] = 'dumb'
        app = Logria(None, False, False)

        # Fake window size: 10 x 100
        app.height = 10
        app.width = 100

        # Set fake previous render
        app.last_row = app.height - 3  # simulate the last row we can render to
        app.current_end = 80  # Simulate the last message rendered

        # Set fake messages
        app.messages = [str(x) for x in range(10)]
        app.parser_index = 0
        app.last_index_processed = 0
        app.analytics_enabled = True

        # Set parser, activate
        app.parser = Parser()
        app.parser.set_pattern(
            pattern=r'(\d)',
            type_='regex',
            name='Test',
            example='4',
            analytics_methods={
                'Item': 'sum'
            }
        )

        # Set analytics method manually
        app.parser._analytics_map = dict(
            zip(range(len(app.parser._analytics_methods.keys())), app.parser._analytics_methods.keys()))

        # Store previous message pointer
        app.previous_messages = app.messages

        # Process parser
        process_parser(app)

        self.assertEqual(app.messages, ['Item', '  Total: 45'])

    def test_process_parser_analytics_invalid(self):
        """
        Test that we correctly process parser with invalid metric
        """
        os.environ['TERM'] = 'dumb'
        app = Logria(None, False, False)

        # Fake window size: 10 x 100
        app.height = 10
        app.width = 100

        # Set fake previous render
        app.last_row = app.height - 3  # simulate the last row we can render to
        app.current_end = 80  # Simulate the last message rendered

        # Set fake messages
        app.messages = [str(x) for x in range(10)]
        app.parser_index = 0
        app.last_index_processed = 0
        app.analytics_enabled = True

        # Set parser, activate
        app.parser = Parser()
        app.parser.set_pattern(
            pattern=r'(\d)',
            type_='fake_type',
            name='Test',
            example='4',
            analytics_methods={
                'Item': 'average'
            }
        )

        # Set analytics method manually
        app.parser._analytics_map = dict(
            zip(range(len(app.parser._analytics_methods.keys())), app.parser._analytics_methods.keys()))

        # Store previous message pointer
        app.previous_messages = app.messages

        # Process parser
        with self.assertRaises(ValueError):
            process_parser(app)

    def test_process_parser_analytics_too_many_matches(self):
        """
        Test that we correctly process parser with too many matches
        """
        os.environ['TERM'] = 'dumb'
        app = Logria(None, False, False)

        # Fake window size: 10 x 100
        app.height = 10
        app.width = 100

        # Set fake previous render
        app.last_row = app.height - 3  # simulate the last row we can render to
        app.current_end = 80  # Simulate the last message rendered

        # Set fake messages
        app.messages = [str(x) for x in range(10, 20)]
        app.parser_index = 0
        app.last_index_processed = 0
        app.analytics_enabled = True

        # Set parser, activate
        app.parser = Parser()
        app.parser.set_pattern(
            pattern=r'(\d)(\d)',
            type_='regex',
            name='Test',
            example='4',
            analytics_methods={
                'Item': 'invalid'
            }
        )

        # Set analytics method manually
        app.parser._analytics_map = dict(
            zip(range(len(app.parser._analytics_methods.keys())), app.parser._analytics_methods.keys()))

        # Store previous message pointer
        app.previous_messages = app.messages

        # Process parser
        process_parser(app)

        self.assertEqual(app.messages, [])

    def test_process_parser_regex_no_pattern(self):
        """
        Test that we correctly process regex parser with no pattern
        """
        os.environ['TERM'] = 'dumb'
        app = Logria(None, False, False)

        # Fake window size: 10 x 100
        app.height = 10
        app.width = 100

        # Set fake previous render
        app.last_row = app.height - 3  # simulate the last row we can render to
        app.current_end = 80  # Simulate the last message rendered

        # Set fake messages
        app.messages = [str(x) for x in range(10)]
        app.parser_index = 0
        app.last_index_processed = 0

        # Set parser, activate
        app.parser = Parser()
        app.parser.set_pattern(
            pattern=None,
            type_='regex',
            name='Test',
            example='4',
            analytics_methods={
                'Item': 'average'
            }
        )

        # Store previous message pointer
        app.previous_messages = app.messages

        # Process parser
        with self.assertRaises(ValueError):
            process_parser(app)

    def test_process_parser_regex(self):
        """
        Test that we correctly process regex parser
        """
        os.environ['TERM'] = 'dumb'
        app = Logria(None, False, False)

        # Fake window size: 10 x 100
        app.height = 10
        app.width = 100

        # Set fake previous render
        app.last_row = app.height - 3  # simulate the last row we can render to
        app.current_end = 80  # Simulate the last message rendered

        # Set fake messages
        app.messages = [str(x) for x in range(10)]
        app.parser_index = 0
        app.last_index_processed = 0

        # Set parser, activate
        app.parser = Parser()
        app.parser.set_pattern(
            pattern=r'(\d)',
            type_='regex',
            name='Test',
            example='4',
            analytics_methods={
                'Item': 'average'
            }
        )

        # Store previous message pointer
        app.previous_messages = app.messages

        # Process parser
        self.assertEqual(app.messages, [str(x) for x in range(10)])

    def test_process_parser_split_no_pattern(self):
        """
        Test that we correctly process parser with split no pattern
        """
        os.environ['TERM'] = 'dumb'
        app = Logria(None, False, False)

        # Fake window size: 10 x 100
        app.height = 10
        app.width = 100

        # Set fake previous render
        app.last_row = app.height - 3  # simulate the last row we can render to
        app.current_end = 80  # Simulate the last message rendered

        # Set fake messages
        app.messages = [str(x) for x in range(10)]
        app.parser_index = 0
        app.last_index_processed = 0

        # Set parser, activate
        app.parser = Parser()
        app.parser.set_pattern(
            pattern=None,
            type_='split',
            name='Test',
            example='4',
            analytics_methods={
                'Item': 'average'
            }
        )

        # Store previous message pointer
        app.previous_messages = app.messages

        # Process parser
        with self.assertRaises(ValueError):
            process_parser(app)

    def test_process_parser_analytics_split(self):
        """
        Test that we correctly process split parsers
        """
        os.environ['TERM'] = 'dumb'
        app = Logria(None, False, False)

        # Fake window size: 10 x 100
        app.height = 10
        app.width = 100

        # Set fake previous render
        app.last_row = app.height - 3  # simulate the last row we can render to
        app.current_end = 80  # Simulate the last message rendered

        # Set fake messages
        app.messages = [f'{x}|{x // 2}' for x in range(10)]
        app.parser_index = 1
        app.last_index_processed = 0

        # Set parser, activate
        app.parser = Parser()
        app.parser.set_pattern(
            pattern=r'\|',
            type_='split',
            name='Test',
            example='4|4',
            analytics_methods={
                'Item 1': 'count',
                'Item 2': 'sum'
            }
        )

        # Store previous message pointer
        app.previous_messages = app.messages

        # Process parser
        process_parser(app)

        self.assertEqual(app.messages, [f'{x // 2}' for x in range(10)])

    def test_process_parser_analytics_average_no_numbers(self):
        """
        Test that we correctly process a parser with average metric but no source numbers
        """
        os.environ['TERM'] = 'dumb'
        app = Logria(None, False, False)

        # Fake window size: 10 x 100
        app.height = 10
        app.width = 100

        # Set fake previous render
        app.last_row = app.height - 3  # simulate the last row we can render to
        app.current_end = 80  # Simulate the last message rendered

        # Set fake messages
        app.messages = [chr(x) for x in range(64, 80)]
        app.parser_index = 0
        app.last_index_processed = 0
        app.analytics_enabled = True

        # Set parser, activate
        app.parser = Parser()
        app.parser.set_pattern(
            pattern=r'(\d)',
            type_='regex',
            name='Test',
            example='4',
            analytics_methods={
                'Item': 'average'
            }
        )

        # Set analytics method manually
        app.parser._analytics_map = dict(
            zip(range(len(app.parser._analytics_methods.keys())), app.parser._analytics_methods.keys()))

        # Since we manually construct alaytics, create the the key
        app.parser.analytics[0] = None
        self.assertIsNone(app.parser.apply_analytics(0, 'A'))



"""
Main app loop, receive CLI args
"""


import argparse
import os
import sys

from logria import APP_NAME, VERSION
from logria.communication.input_handler import CommandInputStream
from logria.communication.shell_output import Logria
from logria.utilities import constants

def main():
    """
    Main app loop, handles parsing args and starting the app
    """
    # Setup CLI args
    parser = argparse.ArgumentParser(
        description=constants.APP_DESCRIPTION,
    )
    parser.add_argument('-v', '--version', action='version',
                        version=f'{APP_NAME} {VERSION}')
    parser.add_argument('-e', action='append', type=str,
                        help=constants.EXEC_HELP)
    parser.add_argument('-c', '--no-cache', dest='no_cache', default=True, action='store_false',
                        help=constants.HISTORY_HELP)
    parser.add_argument('-n', '--no-smart-speed', dest='no_smart_speed', default=True, action='store_false',
                        help=constants.SMART_SPEED_HELP)

    args = parser.parse_args()

    # pylint: disable=no-else-raise
    if not os.isatty(0):
        # If we are getting piped to, use that input stream only
        print(constants.PIPE_INPUT_ERROR)
        sys.exit(-1)
    else:
        if args.e:
            command = args.e[0].split(' ')
            stream = CommandInputStream(command)
            stream.start()
        else:
            # If the stream is None, the app will ask the user to init
            stream = None
        app = Logria(stream, history_tape_cache=args.no_cache, smart_poll_rate=args.no_smart_speed)

    app.start()


if __name__ == '__main__':
    main()



"""
Logria regex handler
"""


import curses

from logria.utilities import constants
from logria.utilities.regex_generator import regex_test_generator
from logria.logger.processor import process_matches

# from logria.communication.shell_output import Logria


def reset_regex_status(logria: 'Logria') -> None:  # type: ignore
    """
    Reset current regex/filter status to no filter
    """
    if logria.parser:
        logria.current_status = f'Parsing with {logria.parser.get_name()}, field {logria.parser_index}'
    else:
        logria.current_status = 'No filter applied'  # CLI message, rendered after
    logria.previous_render = None  # Reset previous render
    logria.func_handle = None  # Disable filter
    logria.highlight_match = False  # Disable highlighting
    logria.regex_pattern = ''  # Clear the current pattern
    logria.matched_rows = []  # Clear out matched rows
    logria.last_index_regexed = 0  # Reset the last searched index
    logria.current_end = 0  # We now do not know where to end
    logria.stick_to_bottom = True  # Stay at the bottom for the next render
    logria.write_to_command_line(logria.current_status)  # Render status


def handle_regex_command(logria: 'Logria', command: str) -> None:  # type: ignore
    """
    Handle a regex command
    """
    reset_regex_status(logria)
    regex_func = regex_test_generator(command)
    if regex_func is None:
        logria.write_to_command_line(f'Invalid regex pattern: {command}')
        return
    logria.func_handle = regex_test_generator(command)
    logria.highlight_match = True
    logria.regex_pattern = command

    # Tell the user what is happening since this is synchronous
    logria.current_status = f'Searching buffer for regex /{logria.regex_pattern}/'
    logria.write_to_command_line(logria.current_status)

    # Process any new matched messages to render
    process_matches(logria)

    # Tell the user we are now filtering
    logria.current_status = f'Regex with pattern /{logria.regex_pattern}/'
    logria.write_to_command_line(logria.current_status)

    # Render the text
    logria.render_text_in_output()
    curses.curs_set(0)


def handle_regex(logria: 'Logria'):  # type: ignore
    """
    Handle when user activates regex mode, including parsing textbox message
    """
    # Handle smart poll rate
    if logria.smart_poll_rate:
        # Make it smooth to type
        logria.update_poll_rate(constants.SLOWEST_POLL_RATE)
    if not logria.analytics_enabled:  # Disable regex in analytics view
        # Handle getting input from the command line for regex
        logria.activate_prompt()
        command = logria.box.gather().strip()
        if command:
            if command == ':q':
                reset_regex_status(logria)
            else:
                handle_regex_command(logria, command)
        else:
            # If command is an empty string, ignore the input
            reset_regex_status(logria)
            logria.reset_command_line()


def toggle_highlight(logria: 'Logria'):  # type: ignore
    """
    Toggle highlighting of search matches
    """
    logria.previous_render = None  # Force render, but defer draw
    if logria.func_handle and logria.highlight_match:
        logria.highlight_match = False
    elif logria.func_handle and not logria.highlight_match:
        logria.highlight_match = True
    else:
        logria.highlight_match = False



"""
Processors functions to filter messages from Logria streams
"""


# from logria.communication.shell_output import Logria


def process_matches(logria: 'Logria') -> None:  # type: ignore
    """
    Process the matches for filtering, should by async but the commented code here
    does not work

    # TODO: Fix this method
    """
    # def add_to_list(result: multiprocessing.Queue, messages: list, last_idx_searched: int, func_handle: Callable):
    #     """
    #     Main loop will create this separate process to find matches while the main loop runs
    #     """
    #     for index, message in range(last_idx_searched, len(messages)):
    #         print(index, message)
    #         if func_handle(message):
    #             result.put(index)
    #         return result

    # result = multiprocessing.Queue()
    # proc = multiprocessing.Process(target=add_to_list, args=(result, logria.messages, logria.last_index_searched, logria.func_handle))
    # proc.start()
    # proc.join()
    # print('done')
    # logria.last_index_searched = len(logria.messages)
    # while not result.empty:
    #     idx = result.get()
    #     print(idx)
    #     logria.matched_rows.append(idx)
    # logria.write_to_prompt('in method')

    # For each message, add its index to the list of matches; this is more efficient than
    # Storing a second copy of each match
    for index in range(logria.last_index_regexed, len(logria.messages)):
        # pylint: disable=not-callable
        if logria.func_handle and logria.func_handle(logria.messages[index]):
            logria.matched_rows.append(index)
    logria.last_index_regexed = len(logria.messages)


def process_parser(logria: 'Logria'):  # type: ignore
    """
    Load parsed messages to new array if we have matches

    # TODO: Same as process_matches
    """
    for index in range(logria.last_index_processed, len(logria.previous_messages)):
        if logria.analytics_enabled:
            logria.parser.handle_analytics_for_message(
                logria.previous_messages[index])
            logria.messages = logria.parser.analytics_to_list()
            # For some reason this isn't switching back
            logria.last_index_processed = len(logria.previous_messages)
        else:
            if logria.messages is not logria.parsed_messages:
                logria.messages = logria.parsed_messages
            match = logria.parser.parse(logria.previous_messages[index])
            if match:
                try:
                    logria.parsed_messages.append(match[logria.parser_index])
                except IndexError:
                    # If there was an error parsing, the message did not match the current pattern
                    pass
            logria.last_index_processed = len(logria.messages)



"""
Unit Tests for color handler
"""


import unittest

from logria.interface import color_handler


class TestColorHandlerSatnitize(unittest.TestCase):
    """
    Test cases to ensure Color Handler will properly handle invalid characters
    """

    def test_sanitize_null(self):
        """
        Test that we remove null bytes from a string
        """
        sample = '\x00test\x00'
        self.assertEqual(color_handler._sanitize(sample), 'test')

    def test_sanitize_no_null(self):
        """
        Test that we just return the clean string
        """
        sample = 'test'
        self.assertEqual(color_handler._sanitize(sample), 'test')

    def test_sanitize_no_color(self):
        """
        Test that we do not remove color bytes from a string
        """
        sample = '\033test\033'
        self.assertEqual(color_handler._sanitize(sample), '\033test\033')


class TestColorHandlerCodes(unittest.TestCase):
    """
    Test that the Color Handler properly resolves color codes
    """

    def test_can_handle_real_color(self):
        """
        Test that we can parse valid colors
        """
        colors = [
            '[30m30[0m'
            '[31m31[0m'
            '[32m32[0m'
            '[33m33[0m'
            '[34m34[0m'
            '[35m35[0m'
            '[36m36[0m'
            '[40m40[0m'
            '[41m41[0m'
            '[42m42[0m'
            '[43m43[0m'
            '[44m44[0m'
            '[45m45[0m'
            '[46m46[0m'
            '[47m47[0m'
            '[90m90[0m'
            '[91m91[0m'
            '[92m92[0m'
            '[93m93[0m'
            '[94m94[0m'
            '[95m95[0m'
            '[96m96[0m'
            '[100m100[0m'
            '[101m101[0m'
            '[102m102[0m'
            '[103m103[0m'
            '[104m104[0m'
            '[105m105[0m'
            '[106m106[0m'
            '[107m107[0m'
        ]
        for color in colors:
            self.assertNotEqual(color_handler._color_str_to_color_pair(color), -1)

    def test_does_not_crash_invalid_color(self):
        """
        Test that we dont crash when given an invalid color
        """
        self.assertEqual(color_handler._color_str_to_color_pair('[108m108[0m'), 101)



"""
Unit Tests for parser
"""

import unittest
from collections import Counter
from os import remove
from pathlib import Path

from logria.logger import parser
from logria.utilities.constants import SAVED_PATTERNS_PATH


class TestPatternSetup(unittest.TestCase):
    """
    Test cases to ensure parser are the expected values
    """

    def test_can_setup_parser(self):
        """
        Ensure we can setup the parser
        """
        p = parser.Parser()
        self.assertIsInstance(p, parser.Parser)

    def test_setup_folder(self):
        """
        Ensure we correctly setup the folder in the expected place
        """
        p = parser.Parser()
        p.setup_folder()
        self.assertEqual(p.folder, Path(Path.home(), SAVED_PATTERNS_PATH))

    def test_get_name(self):
        """
        Test that we correctly get the current filter name
        """
        sample_name = 'test'
        p = parser.Parser()
        log_message = '2005-03-19 15:10:26,773 - simple_example - CRITICAL - critical message 34'
        p.set_pattern(' - ', 'split', sample_name, log_message, {})
        self.assertEqual(p.get_name(), sample_name)

    def test_set_pattern(self):
        """
        Test that we successfully init the data in the parser
        """
        p = parser.Parser()
        log_message = '2005-03-19 15:10:26,773 - simple_example - CRITICAL - critical message 34'
        p.set_pattern(' - ', 'split', 'Hyphen Separated', log_message, {})
        self.assertEqual(p._pattern, r' - ')
        self.assertEqual(p._type, 'split')
        self.assertEqual(p.get_name(), 'Hyphen Separated')
        self.assertEqual(p._example, log_message)
        self.assertEqual(p._analytics_methods, {})

    def test_cannot_set_invalid_pattern(self):
        """
        Test that we successfully init the data in the parser
        """
        p = parser.Parser()
        log_message = '2005-03-19 15:10:26,773 - simple_example - CRITICAL - critical message 34'
        with self.assertRaises(ValueError):
            p.set_pattern('+', 'split', 'Hyphen Separated', log_message, {})


class TestPatternAnalytics(unittest.TestCase):
    """
    Test cases to ensure we can handle analytics properly
    """

    def test_handle_analytics_for_message(self):
        """
        Test that we can handle analytics, this is a wrapper for `apply_analytics`
        """
        p = parser.Parser()
        log_message = '2005-03-19 15:10:26,773 - simple_example - CRITICAL - critical message'
        analytics_methods = {
            "Date": "date",
            "Caller": "count",
            "Level": "count",
            "Message": "sum"
        }
        p.set_pattern(' - ', 'split', 'Hyphen Separated',
                      log_message, analytics_methods)
        p.handle_analytics_for_message(log_message)
        p.handle_analytics_for_message(log_message)
        expected = {0: None, 1: Counter(
            {'simple_example': 2}), 2: Counter({'CRITICAL': 2}), 3: 0}
        self.assertEqual(p.analytics, expected)

    def test_reset_analytics(self):
        """
        Test that we can reset analytics after capturing some
        """
        p = parser.Parser()
        log_message = '2005-03-19 15:10:26,773 - simple_example - CRITICAL - critical message'
        analytics_methods = {
            "Date": "date",
            "Caller": "count",
            "Level": "count",
            "Message": "sum"
        }
        p.set_pattern(' - ', 'split', 'Hyphen Separated',
                      log_message, analytics_methods)
        p.handle_analytics_for_message(log_message)
        p.handle_analytics_for_message(log_message)
        p.reset_analytics()
        expected = {}
        self.assertEqual(p.analytics, expected)

    def test_get_analytics_for_index(self):
        """
        Test that we correctly handle extracting analytics
        """
        p = parser.Parser()
        log_message = '2005-03-19 15:10:26,773 - simple_example - CRITICAL - critical message'
        analytics_methods = {
            "Date": "date",
            "Caller": "count",
            "Level": "count",
            "Message": "sum"
        }
        p.set_pattern(' - ', 'split', 'Hyphen Separated',
                      log_message, analytics_methods)
        for index, name in enumerate(analytics_methods.keys()):
            self.assertEqual(p.get_analytics_for_index(index), name)

    def test_analytics_to_list(self):
        """
        Test that we can get the analytics data as a list
        """
        p = parser.Parser()
        log_message = '2005-03-19 15:10:26,773 - simple_example - CRITICAL - critical message 22'
        analytics_methods = {
            "Date": "date",
            "Caller": "count",
            "Level": "count",
            "Message": "sum"
        }
        p.set_pattern(' - ', 'split', 'Hyphen Separated',
                      log_message, analytics_methods)
        p.handle_analytics_for_message(log_message)
        p.handle_analytics_for_message(log_message)
        expected = ['Caller', '  simple_example: 2', 'Level',
                    '  CRITICAL: 2', 'Message', '  Total: 44']
        self.assertEqual(p.analytics_to_list(), expected)


class TestStringParserMethods(unittest.TestCase):
    """
    Test that we correctly parse strings
    """

    def test_extract_numbers_from_message_int(self):
        """
        Test that we properly parse ints from strings
        """
        p = parser.Parser()
        message = 'abc123a'
        actual = p.extract_numbers_from_message(message)
        self.assertEqual(actual, 123)
        self.assertIsInstance(actual, int)

    def test_extract_numbers_from_message_float(self):
        """
        Test that we properly parse floats from strings
        """
        p = parser.Parser()
        message = 'abc12.3a'
        actual = p.extract_numbers_from_message(message)
        self.assertEqual(actual, 12.3)
        self.assertIsInstance(actual, float)

    def test_clean_ansi_codes(self):
        """
        Test that we successfully remove escape codes from a log
        """
        p = parser.Parser()
        log_message = '\u001b[33m2005-03-19 15:10:26,773 \u001b[0m- \u001b[33msimple_example \u001b[0m- \u001b[33mCRITICAL \u001b[0m- \u001b[33mcritical message 22'
        expected = '2005-03-19 15:10:26,773 - simple_example - CRITICAL - critical message 22'
        self.assertEqual(p.clean_ansi_codes(log_message), expected)

    def test_split_pattern(self):
        """
        Test that we can split a message as we expect
        """
        p = parser.Parser()
        log_message = '2005-03-19 15:10:26,773 - simple_example - CRITICAL - critical message 22'
        analytics_methods = {
            "Date": "date",
            "Caller": "count",
            "Level": "count",
            "Message": "sum"
        }
        p.set_pattern(' - ', 'split', 'Hyphen Separated',
                      log_message, analytics_methods)
        actual = p.split_pattern(log_message)
        expected = [
            '2005-03-19 15:10:26,773',
            'simple_example',
            'CRITICAL',
            'critical message 22'
        ]
        self.assertEqual(actual, expected)

    def test_regex_pattern(self):
        """
        Test that we can regex a message as we expect
        """
        p = parser.Parser()
        log_message = '127.0.0.1 user-identifier frank [10/Oct/2000:13:55:36 -0700] \"GET /apache_pb.gif HTTP/1.0\" 200 2326'
        p.set_pattern("([^ ]*) ([^ ]*) ([^ ]*) \\[([^]]*)\\] \"([^\"]*)\" ([^ ]*) ([^ ]*)", 'regex', 'Common Log Format',
                      log_message, {})
        actual = p.regex_pattern(log_message)
        expected = [
            '127.0.0.1',
            'user-identifier',
            'frank',
            '10/Oct/2000:13:55:36 -0700',
            'GET /apache_pb.gif HTTP/1.0',
            '200',
            '2326'
        ]
        self.assertEqual(actual, expected)

    def test_parse_message_split(self):
        """
        Test that we can split a message through the abstraction
        """
        p = parser.Parser()
        log_message = '2005-03-19 15:10:26,773 - simple_example - CRITICAL - critical message 22'
        analytics_methods = {
            "Date": "date",
            "Caller": "count",
            "Level": "count",
            "Message": "sum"
        }
        p.set_pattern(' - ', 'split', 'Hyphen Separated',
                      log_message, analytics_methods)
        actual = p.parse(log_message)
        expected = [
            '2005-03-19 15:10:26,773',
            'simple_example',
            'CRITICAL',
            'critical message 22'
        ]
        self.assertEqual(actual, expected)

    def test_parse_message_regex(self):
        """
        Test that we can regex a message through the abstraction
        """
        p = parser.Parser()
        log_message = '127.0.0.1 user-identifier frank [10/Oct/2000:13:55:36 -0700] \"GET /apache_pb.gif HTTP/1.0\" 200 2326'
        p.set_pattern("([^ ]*) ([^ ]*) ([^ ]*) \\[([^]]*)\\] \"([^\"]*)\" ([^ ]*) ([^ ]*)", 'regex', 'Common Log Format',
                      log_message, {})
        actual = p.parse(log_message)
        expected = [
            '127.0.0.1',
            'user-identifier',
            'frank',
            '10/Oct/2000:13:55:36 -0700',
            'GET /apache_pb.gif HTTP/1.0',
            '200',
            '2326'
        ]
        self.assertEqual(actual, expected)


class TestPatternInteraction(unittest.TestCase):
    """
    Tests for methods that interact with parser properties
    """

    def test_as_dict(self):
        """
        Test that the dictionary representation is accurate
        """
        p = parser.Parser()
        log_message = '2005-03-19 15:10:26,773 - simple_example - CRITICAL - critical message 22'
        analytics_methods = {
            "Date": "date",
            "Caller": "count",
            "Level": "count",
            "Message": "sum"
        }
        p.set_pattern(' - ', 'split', 'Hyphen Separated',
                      log_message, analytics_methods)
        actual = p.as_dict()
        expected = {
            'pattern': ' - ',
            'type': 'split',
            'name': 'Hyphen Separated',
            'example': '2005-03-19 15:10:26,773 - simple_example - CRITICAL - critical message 22',
            'analytics': {
                'Date': 'date',
                'Caller': 'count',
                'Level': 'count',
                'Message': 'sum'
            }
        }
        self.assertEqual(actual, expected)

    def test_as_list(self):
        """
        Test that the dictionary representation is accurate
        """
        p = parser.Parser()
        log_message = '2005-03-19 15:10:26,773 - simple_example - CRITICAL - critical message 22'
        analytics_methods = {
            "Date": "date",
            "Caller": "count",
            "Level": "count",
            "Message": "sum"
        }
        p.set_pattern(' - ', 'split', 'Hyphen Separated',
                      log_message, analytics_methods)
        actual = p.as_list()
        expected = [
            'Pattern:  - ',
            'Type: split',
            'Name: Hyphen Separated',
            'Example: 2005-03-19 15:10:26,773 - simple_example - CRITICAL - critical message 22',
            "Analytics: {'Date': 'date', 'Caller': 'count', 'Level': 'count', 'Message': 'sum'}"]
        self.assertEqual(actual, expected)

    def test_save_load(self):
        """
        Test that we can successfully save and load items
        """
        p = parser.Parser()
        log_message = '2005-03-19 15:10:26,773 - simple_example - CRITICAL - critical message 22'
        analytics_methods = {
            "Date": "date",
            "Caller": "count",
            "Level": "count",
            "Message": "sum"
        }
        name = 'Test Run'
        p.set_pattern(' - ', 'split', name,
                      log_message, analytics_methods)
        saved = p.as_dict()
        p.save()
        p2 = parser.Parser()
        p2.load(name)
        loaded = p2.as_dict()
        self.assertEqual(loaded, saved)
        remove(Path(p2.folder) / name)

    def test_cannot_load_twice(self):
        """
        Test that we can successfully save and load items
        """
        p = parser.Parser()
        log_message = '2005-03-19 15:10:26,773 - simple_example - CRITICAL - critical message 22'
        analytics_methods = {
            "Date": "date",
            "Caller": "count",
            "Level": "count",
            "Message": "sum"
        }
        name = 'Test Run'
        p.set_pattern(' - ', 'split', name,
                      log_message, analytics_methods)
        saved = p.as_dict()
        p.save()
        p2 = parser.Parser()
        p2.load(name)
        with self.assertRaises(ValueError):
            p2.load(name)
        remove(Path(p2.folder) / name)

    def test_remove_parser(self):
        """
        Test that we can write a session if we need to
        """
        p = parser.Parser()
        log_message = '2005-03-19 15:10:26,773 - simple_example - CRITICAL - critical message 22'
        analytics_methods = {
            "Date": "date",
            "Caller": "count",
            "Level": "count",
            "Message": "sum"
        }
        name = 'Test Run'
        p.set_pattern(' - ', 'split', name,
                      log_message, analytics_methods)
        saved = p.as_dict()
        p.save()
        p2 = parser.Parser()
        p2.load(name)
        p2.remove('Test Run')
        self.assertNotIn('Test Run', p2.patterns().values())

    def test_cannot_remove_parser(self):
        """
        Test that we can write a session if we need to
        """
        p = parser.Parser()
        self.assertFalse(p.remove('This Fake Parser Name'))

    def test_display_example(self):
        """
        Test that we properly format the example parsed message
        """
        p = parser.Parser()
        log_message = '2005-03-19 15:10:26,773 - simple_example - CRITICAL - critical message 22'
        analytics_methods = {
            "Date": "date",
            "Caller": "count",
            "Level": "count",
            "Message": "sum"
        }
        p.set_pattern(' - ', 'split', 'Hyphen Separated',
                      log_message, analytics_methods)
        actual = p.display_example()
        expected = [
            '0: 2005-03-19 15:10:26,773',
            '1: simple_example',
            '2: CRITICAL',
            '3: critical message 22'
        ]
        self.assertEqual(actual, expected)

    def test_cannot_display_example(self):
        """
        Test that we properly format the example parsed message
        """
        p = parser.Parser()
        with self.assertRaises(ValueError):
            p.display_example()

    def test_patterns(self):
        """
        Test that we read patterns as dict
        """
        p = parser.Parser()
        self.assertIsInstance(p.patterns(), dict)

    def test_show_patterns(self):
        """
        Test that we read patterns as list
        """
        p = parser.Parser()
        self.assertIsInstance(p.show_patterns(), list)



"""
Class to handle keeping track of input history
"""
import os
from pathlib import Path
from typing import List

from logria.utilities.constants import HISTORY_TAPE_NAME, SAVED_HISTORY_PATH, HISTORY_EXCLUDES, LOGRIA_ROOT, USER_HOME


class HistoryTape():
    """
    Class to keep track of user's input history
    """

    def __init__(self, use_cache: bool = True):
        self.history_tape: List[str] = []  # Not a real queue
        self.current_index: int = -1  # The index we are at in the tape
        self.should_scroll_back: bool = False
        self.history_tape_file = f'{SAVED_HISTORY_PATH}/{HISTORY_TAPE_NAME}'
        self.use_cache = use_cache
        if self.use_cache:
            self.load_history_from_file()

    def load_history_from_file(self) -> None:
        """
        Load the history tape from the disk if it exists
        """
        # Create Logria root folder if it is missing
        if Path(f'{USER_HOME}/{LOGRIA_ROOT}').exists():
            pass
        else:
            os.mkdir(Path(f'{USER_HOME}/{LOGRIA_ROOT}'))

        # Create history folder if it is missing
        if Path(SAVED_HISTORY_PATH).exists():
            pass
        else:
            os.mkdir(Path(SAVED_HISTORY_PATH))

        # Read the file into memory
        if os.path.isfile(self.history_tape_file):
            with open(self.history_tape_file) as history_cache:
                self.history_tape = [line.rstrip(
                    '\n') for line in history_cache]
                self.current_index = len(self.history_tape) - 1
        else:
            with open(self.history_tape_file, 'w+') as history_cache:
                history_cache.write('')

    def write_to_history_file(self, message) -> None:
        """
        Write a new message to the history tape
        """
        if Path(SAVED_HISTORY_PATH).exists():
            if os.path.isfile(self.history_tape_file):
                with open(self.history_tape_file, 'a') as history_cache:
                    history_cache.write(message)
                    history_cache.write('\n')

    def size(self) -> int:
        """
        Return the number of items in the tape
        """
        return len(self.history_tape)

    def get_current_item(self) -> str:
        """
        Get the most recent item
        """
        if not self.history_tape:
            return ''
        return self.history_tape[self.current_index]

    def go_to(self, index: int) -> str:
        """
        Go to a specific point in time, bound by the size of the tape
        """
        if index < 0:
            self.current_index = 0
        elif index < len(self.history_tape) - 1:
            self.current_index = index
        elif index >= len(self.history_tape):
            self.current_index = len(self.history_tape) - 1
        return self.get_current_item()

    def tail(self, last_n: int = 5) -> List[str]:
        """
        Get the most recent `last n` messages from the tape
        """
        if last_n < 0:
            return []
        return self.history_tape[-last_n:]

    def at_end(self) -> bool:
        """
        Determine if we are at the end of the list
        """
        return self.current_index == len(self.history_tape) - 1

    def add_item(self, item: str) -> None:
        """
        Adds `item` to the end of the queue and update the index
        """
        clean_item = item.strip()  # Sanitize extra spaces from `gather()`
        if clean_item not in HISTORY_EXCLUDES:
            # history_tape must exist and the last stored item not be the new item
            if not self.history_tape or clean_item != self.history_tape[-1]:
                if self.use_cache:
                    self.write_to_history_file(clean_item)
                self.history_tape.append(clean_item)
                self.current_index = len(self.history_tape) - 1
                self.should_scroll_back = False

    def scroll_back_n(self, num_to_scroll: int) -> str:
        """
        Scroll back one item in the tape
        """
        if self.history_tape:
            if self.should_scroll_back:
                self.current_index = max(0, self.current_index - num_to_scroll)
            else:
                self.should_scroll_back = True
        return self.get_current_item()

    def scroll_forward_n(self, num_to_scroll: int) -> str:
        """
        Scroll forward one item in the tape; if we are at the end return nothing
        """
        if self.at_end():
            return ''
        if self.history_tape:
            self.current_index = min(
                len(self.history_tape) - 1, self.current_index + num_to_scroll)
        return self.get_current_item()

    def scroll_back(self) -> str:
        """
        Common case where we scroll back a single item
        """
        return self.scroll_back_n(1)

    def scroll_forward(self) -> str:
        """
        Common case where we scroll back a single item
        """
        return self.scroll_forward_n(1)

    def __repr__(self):
        return f'<History Tape containing {self.size():,} items, currently selected {self.get_current_item()}>'



"""
Unit Tests for history tape
"""

import unittest
import uuid

from logria.communication import input_history


class TestHistoryTapeFunctions(unittest.TestCase):
    """
    Test cases to ensure history tape are the expected values
    """

    def test_can_create_history_tape(self):
        """
        Test that we can create a history tape instance without a cache
        """
        tape = input_history.HistoryTape(use_cache=False)
        self.assertIsInstance(tape, input_history.HistoryTape)

    def test_can_create_cache_history_tape(self):
        """
        Test that we can create a history tape instance with a cache
        """
        unique_str = str(uuid.uuid4())
        tape = input_history.HistoryTape()
        tape.add_item(unique_str)
        self.assertIsInstance(tape, input_history.HistoryTape)
        self.assertEqual(tape.get_current_item(), unique_str)

    def test_can_print_history_tape(self):
        """
        Test that we can create a history tape instance without a cache
        """
        tape = input_history.HistoryTape(use_cache=False)
        print(tape)

    def test_get_current_item_empty(self):
        """
        Test that we don't crash when looking at a blank tape
        """
        tape = input_history.HistoryTape(use_cache=False)
        self.assertEqual(tape.get_current_item(), '')

    def test_can_add_item(self):
        """
        Test that we can add and retrieve the current item
        """
        tape = input_history.HistoryTape(use_cache=False)
        tape.add_item('Test')
        print(tape.get_current_item())
        self.assertEqual(tape.get_current_item(), 'Test')

    def test_cant_add_excluded_item(self):
        """
        Test that we don't add restricted items
        """
        tape = input_history.HistoryTape(use_cache=False)
        tape.add_item(':history')
        tape.add_item(':history off')
        self.assertEqual(tape.size(), 0)

    def test_at_end(self):
        """
        Test that we correctly report when we are at the end
        """
        tape = input_history.HistoryTape(use_cache=False)
        tape.add_item('Test')
        self.assertTrue(tape.at_end())

    def test_not_at_end(self):
        """
        Test that we correctly report when we are not at the end
        """
        tape = input_history.HistoryTape(use_cache=False)
        tape.add_item('Test 1')
        tape.add_item('Test 2')
        tape.add_item('Test 3')
        tape.scroll_back()
        tape.scroll_back()
        self.assertFalse(tape.at_end())

    def test_cant_add_item_twice(self):
        """
        Test that we can add and retrieve the current item
        """
        tape = input_history.HistoryTape(use_cache=False)
        tape.add_item('Test')
        tape.add_item('Test')
        self.assertEqual(tape.size(), 1)

    def test_can_scroll_backward_in_bounds(self):
        """
        Ensure we can scroll back
        """
        tape = input_history.HistoryTape(use_cache=False)
        tape.add_item('Test 1')
        tape.add_item('Test 2')
        tape.add_item('Test 3')
        tape.scroll_back()
        tape.scroll_back()
        self.assertEqual(tape.get_current_item(), 'Test 2')

    def test_can_scroll_backward_out_of_bounds(self):
        """
        Ensure we can scroll back
        """
        tape = input_history.HistoryTape(use_cache=False)
        tape.add_item('Test 1')
        tape.add_item('Test 2')
        tape.scroll_back()
        tape.scroll_back()
        tape.scroll_back()
        tape.scroll_back()
        self.assertEqual(tape.get_current_item(), 'Test 1')

    def test_can_scroll_forward_in_bounds_(self):
        """
        Ensure we can scroll forward
        """
        tape = input_history.HistoryTape(use_cache=False)
        tape.add_item('Test 1')
        tape.add_item('Test 2')
        tape.add_item('Test 3')
        tape.scroll_back()
        tape.scroll_back()
        tape.scroll_forward()
        self.assertEqual(tape.get_current_item(), 'Test 3')

    def test_can_scroll_forward_in_bounds(self):
        """
        Ensure we can scroll forward
        """
        tape = input_history.HistoryTape(use_cache=False)
        tape.add_item('Test 1')
        tape.add_item('Test 2')
        tape.add_item('Test 3')
        tape.scroll_back()
        tape.scroll_back()
        tape.scroll_forward()
        tape.scroll_forward()
        tape.scroll_forward()
        tape.scroll_forward()
        tape.scroll_forward()
        self.assertEqual(tape.get_current_item(), 'Test 3')

    def test_can_go_to_in_bounds(self):
        """
        Ensure we can go to a point in the tape in bounds
        """
        tape = input_history.HistoryTape(use_cache=False)
        tape.add_item('Test 1')
        tape.add_item('Test 2')
        tape.add_item('Test 3')
        tape.go_to(0)
        self.assertEqual(tape.get_current_item(), 'Test 1')

    def test_can_go_to_out_of_bounds_up(self):
        """
        Ensure we do not try to go to a point above the tape and instead go to the last item
        """
        tape = input_history.HistoryTape(use_cache=False)
        tape.add_item('Test 1')
        tape.add_item('Test 2')
        tape.add_item('Test 3')
        tape.go_to(10)
        self.assertEqual(tape.get_current_item(), 'Test 3')

    def test_can_go_to_out_of_bounds_down(self):
        """
        Ensure we do not try to go to a point below the tape and instead go to the first item
        """
        tape = input_history.HistoryTape(use_cache=False)
        tape.add_item('Test 1')
        tape.add_item('Test 2')
        tape.add_item('Test 3')
        tape.go_to(-3)
        self.assertEqual(tape.get_current_item(), 'Test 1')

    def test_tape_size(self):
        """
        Test that we can get size of the tape
        """
        tape = input_history.HistoryTape(use_cache=False)
        self.assertEqual(tape.size(), 0)
        for item in range(10):
            tape.add_item(f'Test {item}')
        self.assertEqual(tape.size(), 10)

    def test_tape_tail(self):
        """
        Test that we can tail the tape
        """
        tape = input_history.HistoryTape(use_cache=False)
        for item in range(10):
            tape.add_item(f'Test {item}')
        self.assertEqual(
            tape.tail(4), ['Test 6', 'Test 7', 'Test 8', 'Test 9'])

    def test_tape_tail_bad_index(self):
        """
        Test that we don't crash tail with invalid indexes
        """
        tape = input_history.HistoryTape(use_cache=False)
        for item in range(10):
            tape.add_item(f'Test {item}')
        self.assertEqual(tape.tail(-1), [])

    def test_can_tail_default_n(self):
        """
        Test that we can get the last 5 messages
        """
        tape = input_history.HistoryTape(use_cache=False)
        num_to_get = 5
        for item in range(10):
            tape.add_item(f'Test {item}')
        self.assertEqual(
            tape.tail(),
            [f'Test {x + (tape.size() - num_to_get)}' for x in range(5)]
        )

    def test_can_tail_custom_end(self):
        """
        Test that we can get the last n messages
        """
        tape = input_history.HistoryTape(use_cache=False)
        num_to_get = 7
        for item in range(10):
            tape.add_item(f'Test {item}')
        self.assertEqual(
            tape.tail(num_to_get),
            [f'Test {x + (tape.size() - num_to_get)}' for x in range(num_to_get)]
        )



"""
Unit Tests for regex_generator
"""

import unittest

from logria.utilities import regex_generator


class TestRegexGenerator(unittest.TestCase):
    """
    Test cases to ensure regex_generator is handling data properly
    """

    def test_validator(self):
        """
        Test that we correctly generate a pattern
        """
        self.assertTrue(callable(regex_generator.regex_test_generator('-')))

    def test_generated_regex_func(self):
        """
        Test that we properly generate a regex function
        """
        pattern = regex_generator.regex_test_generator('-')
        self.assertTrue(pattern('-'))

    def test_generated_regex_func_with_color_code(self):
        """
        Test that we match properly against a string with escape codes
        """
        pattern = regex_generator.regex_test_generator(' - ')
        self.assertTrue(pattern(' \u001b[0m-\u001b[32m '))

    def test_generated_invalid_regex(self):
        """
        Test that we match properly against a string with escape codes
        """
        pattern = regex_generator.regex_test_generator('(')
        self.assertIsNone(pattern)


class TestRealLength(unittest.TestCase):
    """
    Test that we properly get real lengths
    """

    def test_get_real_length(self):
        """
        Test that we get the real length of a string
        """
        self.assertEqual(regex_generator.get_real_length('word'), 4)

    def test_get_real_length_with_color_code(self):
        """
        Test that we get the real length of a string with escape codes
        """
        self.assertEqual(regex_generator.get_real_length(
            '\u001b[0m word \u001b[32m'), 6)



"""
Class to handle parsing of standard log patterns
"""

import json
import os
import re
from collections import Counter
from pathlib import Path
from typing import List, Union, Optional

from logria.utilities import fs
from logria.utilities.constants import ANSI_COLOR_PATTERN, SAVED_PATTERNS_PATH


class Parser():
    """
    Handles setting up of log message parsing
    """

    def __init__(self, num_to_print=5):
        self._pattern: Optional[str] = None  # The raw pattern
        # The type of pattern to parse, sring {'split', 'regex'}
        self._type: Optional[str] = None
        self._name: Optional[str] = None  # The name of the pattern
        self._example: Optional[str] = None  # An example used to list on the frontend
        # Analytics methods to use when parsing
        self._analytics_methods: Optional[dict] = None
        # Stores the map of the message index to the analytics method names
        self.num_to_print = num_to_print  # Number of items to print in analytics output
        self._analytics_map: dict = {}
        self.analytics: dict = {}  # Analytics the main script can access
        self.setup_folder()

    def setup_folder(self):
        """
        Set workspace folder, create if nonexistent
        """
        if Path(SAVED_PATTERNS_PATH).exists():
            pass
        else:
            os.mkdir(Path(SAVED_PATTERNS_PATH))
        self.folder = Path(SAVED_PATTERNS_PATH)

    def get_name(self):
        """
        Get the name of the parser as a string
        """
        return self._name

    def set_pattern(self, pattern: str, type_: str, name: str, example: str, analytics_methods: dict) -> None:
        """
        Init the class variables when loading
        """
        if pattern is not None:
            try:
                re.compile(pattern)
            except re.error as err:
                raise ValueError(
                    f'Parser {name} has invalid regex pattern: /{pattern}/') from err
        self._pattern = pattern
        self._type = type_
        self._name = name
        self._example = example
        self._analytics_methods = analytics_methods
        self._analytics_map = dict(
            zip(range(len(analytics_methods.keys())), analytics_methods.keys()))

    def reset_analytics(self) -> None:
        """
        Resets the current analytics dictionary
        """
        self.analytics = {}

    def get_analytics_for_index(self, index: int) -> str:
        """
        Returns the name of the item that is the provided index
        """
        return self._analytics_map[index]

    # pylint: disable=no-self-use
    def extract_numbers_from_message(self, message: str) -> Union[int, float]:
        r"""
        We do not use regex replacement here because...

        % python -m timeit '"".join(c for c in "sdkjh987978asd098as0980a98sd" if c.isdigit() or c == ".")'
        100000 loops, best of 3: 3.47 usec per loop
        % python -m timeit 'import re; re.sub(r"[^\d\.]", "", "sdkjh987978asd098as0980a98sd")'
        100000 loops, best of 3: 4.67 usec per loop
        """
        digits = "".join(c for c in message if c.isdigit() or c == ".")
        out_num = float(digits)
        if out_num.is_integer():
            return int(out_num)
        return out_num

    def apply_analytics(self, index: int, part: str) -> None:
        """
        Applies an analytics rule to a message
        """
        # Figure out what rule we want to apply
        rule_name = self.get_analytics_for_index(index)
        if not self._analytics_methods:
            return None
        rule = self._analytics_methods[rule_name]
        if rule == 'count':
            if not self.analytics[index]:
                self.analytics[index] = Counter()
            self.analytics[index].update([part])
        elif rule == 'sum':
            if not self.analytics[index]:
                self.analytics[index] = 0
            try:
                val = self.extract_numbers_from_message(part)
                self.analytics[index] += val
            except ValueError:
                pass
        elif rule == 'average':
            if not self.analytics[index]:
                self.analytics[index] = {'average': 0, 'count': 0, 'total': 0}
            try:
                val = self.extract_numbers_from_message(part)
                self.analytics[index]['count'] += 1
                self.analytics[index]['total'] += val
                self.analytics[index]['average'] = self.analytics[index]['total'] / \
                    self.analytics[index]['count']
            except ValueError:
                pass
        return None

    def handle_analytics_for_message(self, message: str) -> None:
        """
        Applies the analytics rules for each part of a message
        """
        if message:
            parsed_message = self.parse(message)
            if parsed_message is not None:
                for index, part in enumerate(parsed_message):
                    if index not in self.analytics:
                        self.analytics[index] = None
                    try:
                        self.apply_analytics(index, part)
                    except KeyError:
                        # If we matched too many fields, this message is invalid
                        pass

    def analytics_to_list(self) -> List[str]:
        """
        Return the existing parsed analytics as a list
        """
        out_l = []
        for stat in self.analytics:
            value = self.analytics[stat]
            if value is None:
                continue
            out_l += [f'{self.get_analytics_for_index(stat)}']
            if isinstance(value, Counter):
                out_l += [f'  {item}: {count:,}' for item,
                          count in value.most_common(self.num_to_print)]
            elif isinstance(value, (int, float)):
                out_l += [f'  Total: {value:,}']
            elif isinstance(value, dict):
                out_l += [f'  {key}:\t {value[key]:,.2f}' for key in value]
        return out_l

    def clean_ansi_codes(self, string: str) -> str:
        """
        Remove ANSI escape sequences from a string
        """
        return re.sub(ANSI_COLOR_PATTERN, '', string)

    def split_pattern(self, message: str) -> List[str]:
        """
        Split a log message based on a delimiter
        """
        if self._pattern is None:
            raise ValueError('Parsing pattern when pattern not set!')
        parts = re.split(self._pattern, message)
        return parts

    def regex_pattern(self, message: str) -> List[str]:
        """
        Split a log message based on matches to a regex pattern
        """
        if self._pattern is None:
            raise ValueError('Parsing pattern when pattern not set!')
        matches = re.match(self._pattern, message)
        return list(matches.groups()) if matches is not None else []

    def parse(self, message: str) -> List[str]:
        """
        Parse a log message
        """
        if self._type == 'split':
            return self.split_pattern(message)
        elif self._type == 'regex':
            return self.regex_pattern(message)
        raise ValueError(f'{self._type} is not a valid Parser type!')

    def as_dict(self):
        """
        Dict representation
        """
        return {'pattern': self._pattern,
                'type': self._type,
                'name': self._name,
                'example': self._example,
                'analytics': self._analytics_methods
                }

    def as_list(self):
        """
        List representation
        """
        return [
            f'Pattern: {self._pattern}',
            f'Type: {self._type}',
            f'Name: {self._name}',
            f'Example: {self._example}',
            f'Analytics: {self._analytics_methods}'
        ]

    def save(self) -> None:
        """
        Save current pattern
        """
        # Convert to json
        if self._pattern:
            d = self.as_dict()
            with open(self.folder / self._name, 'w') as f:
                f.write(json.dumps(d, indent=4))

    def load(self, name: str) -> None:
        """
        Load existing pattern
        """
        # Convert to json
        if self._pattern is not None:
            raise ValueError('Setting pattern while pattern already set!')
        patterns = set(fs.listdir(self.folder, {'.DS_Store'}))
        if name in patterns:
            with open(self.folder / name, 'r') as f:
                d = json.loads(f.read())
                self.set_pattern(d['pattern'], d['type'],
                                 d['name'], d['example'], d['analytics'])
                self._analytics_map = dict(
                    zip(range(len(d['analytics'].keys())), d['analytics'].keys()))

    def remove(self, parser_name: str) -> bool:
        """
        Remove a parser from the list of parser

        parser_name is the filename of the parser
        """
        path = self.folder / parser_name
        if os.path.exists(path):
            os.remove(path)
            return True
        return False  # If we were not able to delete the file

    def display_example(self):
        """
        Show parser result so the user can choose what section to look at
        """
        if self._pattern is None:
            raise ValueError('Display called without loading a parser!')
        match = self.parse(self._example)
        return [f'{i}: {v}' for i, v in enumerate(match)]

    def patterns(self) -> dict:
        """
        Get the existing patterns as a dict
        """
        patterns = fs.listdir(self.folder, {'.DS_Store'})
        return dict(zip(range(0, len(patterns)), patterns))

    def show_patterns(self) -> List[str]:
        """
        Get the existing patterns as a list
        """
        patterns = fs.listdir(self.folder, {'.DS_Store'})
        return [f'{i}: {v}' for i, v in enumerate(patterns)]



"""
Utilities to handle input validation
"""


from typing import Callable, Dict

from logria.commands import command, edit, parser, regex, scroll, window

STROKES: Dict[str, Callable] = {
    '/': regex.handle_regex,
    'h': regex.toggle_highlight,
    ':': command.handle_command,
    'i': edit.toggle_insert_mode,
    's': window.swap_input,
    'p': parser.enable_parser,
    'a': parser.enable_analytics,
    'z': parser.teardown_parser,
    'KEY_RESIZE': window.resize,
    'KEY_UP': scroll.up,
    'KEY_DOWN': scroll.down,
    'KEY_PPAGE': scroll.pgup,
    'KEY_NPAGE': scroll.pgdn,
    'KEY_RIGHT': scroll.bottom,
    'KEY_LEFT': scroll.top,
}


def resolve_keypress(logria: 'Logria', key_press: str) -> None:  # type: ignore
    """
    Resolve a key press to a callable in the dictionary
    """
    if key_press in STROKES:
        STROKES[key_press](logria)


def validator(key: int):
    """
    Intercept invalid strokes and convert them to valid ones
    """
    # Handle backspace on MacOS
    if key == 127:
        return 263  # Ctrl-h
    return key



"""
Functions to handler rendering content in the Logria window
"""


from math import ceil
from typing import List, Tuple

from logria.utilities.regex_generator import get_real_length

# from logria.communication.shell_output import Logria


def determine_position(logria: 'Logria', messages_pointer: List[str]) -> Tuple[int, int]:  # type: ignore
    """
    Determine the start and end positions for a screen render
    """
    if logria.stick_to_top:
        end = 0
        rows = 0
        for i in messages_pointer:
            if messages_pointer is logria.messages:
                # No processing needed for normal messages
                item: str = i
            elif messages_pointer is logria.matched_rows:
                # Grab the matched message
                item = logria.messages[i]  # type: ignore
            # Determine if the message will fit in the window
            msg_lines = ceil(get_real_length(item) / logria.width)
            rows += msg_lines
            # If we can fit, increment the last row number
            if rows < logria.last_row and end < len(messages_pointer) - 1:
                end += 1
            else:
                break
        logria.current_end = end  # Save this row so we know where we are
        # When iterating backwards, we need to end at 0, so we must create a range
        # object like range(10, -1, -1) to generate a list that ends at 0
        # If there are no messages, we want to not iterate later, so we change the
        # -1 to 0 so that we do not iterate at all
        return -1 if messages_pointer else 0, end  # Early escape
    elif logria.stick_to_bottom:
        end = len(messages_pointer) - 1
    elif logria.manually_controlled_line:
        if len(messages_pointer) < logria.last_row:
            # If have fewer messages than lines, just render it all
            end = len(messages_pointer) - 1
        elif logria.current_end < logria.last_row:
            # If the last row we rendered comes before the last row we can render,
            # use all of the available rows
            end = logria.current_end
        elif logria.current_end < len(messages_pointer):
            # If we are looking at a valid line, render ends there
            end = logria.current_end
        else:
            # If we have over-scrolled, go back
            if logria.current_end > len(messages_pointer):
                logria.current_end = len(messages_pointer)
            # Since current_end can be zero, we have to use the number of messages
            end = len(messages_pointer)
    else:
        end = len(messages_pointer)
    logria.current_end = end  # Save this row so we know where we are
    # Last index of a list is length - 1
    start = max(-1, end - logria.last_row - 1)
    return start, end



"""
Unit Tests for constants
"""

import unittest

from logria.utilities import constants


class TestFormattingFunctions(unittest.TestCase):
    """
    Test cases to ensure constants are the expected values
    """

    def test_ansi_color_pattern(self):
        """
        Ensure ANSI_COLOR_PATTERN is the expected value
        """
        self.assertEqual(constants.ANSI_COLOR_PATTERN,
                         r'(\x9B|\x1B\[)[0-?]*[ -\/]*[@-~]')

    def test_root_folder(self):
        """
        Ensure LOGRIA_ROOT is the expected value
        """
        self.assertEqual(constants.LOGRIA_ROOT, '.logria')

    def test_saved_patterns_path(self):
        """
        Ensure SAVED_PATTERNS_PATH is the expected value
        """
        self.assertEqual(constants.SAVED_PATTERNS_PATH,
                         f'{constants.USER_HOME}/.logria/patterns')

    def test_saved_sessions_path(self):
        """
        Ensure SAVED_SESSIONS_PATH is the expected value
        """
        self.assertEqual(constants.SAVED_SESSIONS_PATH,
                         f'{constants.USER_HOME}/.logria/sessions')



"""
Contains utilities and constants that control how the app handles regex
"""

import re
from typing import Callable, Optional

from logria.utilities.constants import ANSI_COLOR_PATTERN


def regex_test_generator(pattern: str) -> Optional[Callable]:
    """
    Return a function that will test a string against `pattern`
    Ignores characters in ANSI color escape codes
    """
    try:
        re.compile(pattern)
    except re.error:
        return None
    return lambda string: bool(re.search(pattern,
                                         re.sub(ANSI_COLOR_PATTERN,
                                                '',
                                                string)))


def get_real_length(item: str) -> int:
    """
    Get the real length of a string without escape codes
    """
    return len(re.sub(ANSI_COLOR_PATTERN, '', item))



"""
Filesystem interaction functions
"""


import os
from pathlib import Path
from typing import Union, List


def listdir(path: Union[str, Path], ignored_patterns: set) -> List:
    """
    Safe os.listdir wrapper that ignores patterns in filenames
    """
    if os.path.exists(path) and os.path.isdir(path):
        return [x for x in os.listdir(path) if x not in ignored_patterns]
    return []



from setuptools import setup, find_packages
from logria import APP_NAME, VERSION

with open("README.md", "r") as fh:
    LONG_DESCRIPTION = fh.read()

setup(
    name=APP_NAME,
    version=VERSION,
    description='A powerful CLI tool that puts log analytics at your fingertips.',
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    author='Christopher Sardegna',
    author_email='logria@reagentx.net',
    url='https://github.com/ReagentX/Logria-py',
    packages=find_packages(),
    entry_points={
        'console_scripts': [
            'logria = logria.__main__:main'
        ]
    },
    python_requires='>=3.6',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console :: Curses',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.8',
        'Topic :: System :: Logging',
        'Topic :: Internet :: Log Analysis',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ]
)



"""
Handles the internal logger for debugging purposes
"""

import logging


def setup_default_logger(name: str) -> logging.Logger:
    """
    Create a logger instance we can import to anywhere, logs in format:
    2019-10-28 14:32:57,843 - model.views - INFO - Received request to get_data

    To create an instance of the logger, import with
        from repricing_modules.helpers import logger
    and declare with this call:
        LOGGER = logger.setup_default_logger(__name__)
    """
    # Create Python/Django logger instance for the module
    logger = logging.getLogger(name)
    # Log all types of messages
    logger.setLevel(logging.DEBUG)
    # create console handler and set level to debug
    console_handler = logging.StreamHandler()
    console_handler.setLevel(logging.DEBUG)
    # create formatter; use color escape codes
    formatter = logging.Formatter(
        '\u001b[33m%(asctime)s \u001b[0m- \u001b[34m%(name)s.%(funcName)s \u001b[0m- \u001b[32m%(processName)s \u001b[0m- \u001b[36m%(levelname)s \u001b[0m- %(message)s')
    # add formatter to console handler
    console_handler.setFormatter(formatter)
    # add console handler to logger
    logger.addHandler(console_handler)
    return logger



"""
Classes to handle saving sessions
"""


import json
import os
from pathlib import Path
from typing import List, Union

from logria.utilities import fs
from logria.utilities.constants import LOGRIA_ROOT, SAVED_SESSIONS_PATH


class SessionHandler():
    """
    Class that manages sessions for the user
    """

    def __init__(self):
        self._commands: List[List[str]] = []
        self._type: str = ''  # One of {'command', 'file'}
        self.folder: Path = self.setup_folder()

    def set_params(self, command: List[str], type_: str) -> None:
        """
        Update the commands and parameters
        """
        self.add_command(command)
        self.set_type(type_)

    def set_type(self, type_str):
        """
        Sets the type of command
        """
        self._type = type_str

    def add_command(self, command: List[str]):
        """
        Adds the command to the list of commands
        """
        self._commands.append(command)

    # pylint: disable=no-self-use
    def setup_folder(self) -> Path:
        """
        Set workspace folder, create if nonexistent
        """
        home = str(Path.home())
        if Path(home, LOGRIA_ROOT).exists():
            pass
        else:
            os.mkdir(Path(home, LOGRIA_ROOT))
        if Path(home, SAVED_SESSIONS_PATH).exists():
            pass
        else:
            os.mkdir(Path(home, SAVED_SESSIONS_PATH))
        return Path(home, SAVED_SESSIONS_PATH)

    def load_session(self, item: int) -> dict:
        """
        Load data for a session
        """
        out_d = {}
        if item in SessionHandler().sessions():
            name = SessionHandler().sessions()[item]
            with open(self.folder / name, 'r') as f:
                out_d = json.loads(f.read())
        return out_d

    def remove_session(self, session_name: str) -> bool:
        """
        Remove a session from the list of sessions

        session_name is the filename of the session
        """
        path = self.folder / session_name
        if os.path.exists(path):
            os.remove(path)
            return True
        return False  # If we were not able to delete the file

    def save_current_session(self, name: str) -> None:
        """
        Saves the current session
        """
        self.save_session(name, self._commands, self._type)

    def save_session(self, name: str, commands: Union[List[List[str]], List[str]], type_: str) -> None:
        """
        Save a session to the sessions directory
        """
        out_json = {'commands': commands, 'type': type_}
        with open(self.folder / name, 'w') as f:
            f.write(json.dumps(out_json, indent=4))

    def as_list(self) -> List[str]:
        """
        Returns a list representation of the current session
        """
        out_l = []
        out_l.append('Type:')
        out_l.append(f'  {self._type}')
        if self._type == 'command':
            out_l.append('Commands:')
            for command in self._commands:
                out_l.append(f'  {" ".join(command)}')
        elif self._type == 'file':
            out_l.append('Files:')
            for file in self._commands:
                out_l.append(f'  {"/".join(file)}')
        return out_l

    def sessions(self) -> dict:
        """
        Get the existing sessions as a dict
        """
        sessions = sorted(fs.listdir(self.folder, {'.DS_Store'}))
        return dict(zip(range(0, len(sessions)), sessions))

    def show_sessions(self) -> List[str]:
        """
        Get the existing sessions as a list for output
        """
        sessions = sorted(fs.listdir(self.folder, {'.DS_Store'}))
        return [f'{i}: {v}' for i, v in enumerate(sessions)]



"""
Unit Tests for input_handler
"""

import unittest

from logria.communication import input_handler


class TestInputStream(unittest.TestCase):
    """
    Test cases to ensure InputStream can be initialized
    """

    def test_create_input_stream(self):
        """
        Test that we can crate an InputStream
        """
        i = input_handler.InputStream(['ls', '-l'])
        self.assertIsInstance(i, input_handler.InputStream)

    def test_cannot_start_process(self):
        """
        Test that parent InputStream cannot start processes
        Unit tests cannot be pickled, so this test case will never work
        """
        # TODO: Make this test do something
        # i = input_handler.InputStream(['ls', '-l'])
        # with self.assertRaises(NotImplementedError):
        #     i.start()
        pass

    def test_can_exit(self):
        """
        Test that parent InputStream can exit processes
        Unit tests cannot be pickled, so this test case will never work
        """
        i = input_handler.InputStream(['ls', '-l'])
        with self.assertRaises(AttributeError):
            i.exit()


class TestCommandInputStream(unittest.TestCase):
    """
    Test cases to ensure CommandInputStream can be initialized
    """

    def test_create_command_input_stream(self):
        """
        Test that we can create a CommandInputStream
        """
        i = input_handler.CommandInputStream(['ls', '-l'])
        self.assertIsInstance(i, input_handler.CommandInputStream)


class TestFileInputStream(unittest.TestCase):
    """
    Test cases to ensure FileInputStream can be initialized
    """

    def test_create_file_input_stream(self):
        """
        Test that we can create a FileInputStream
        """
        i = input_handler.FileInputStream(['ls', '-l'])
        self.assertIsInstance(i, input_handler.FileInputStream)
